{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances -fallow-undecidable-instances #-}
-----------------------------------------------------------------------------
-- Maintainer  :  Alexandra and Joost Visser
-- Stability   :  experimental
-- Portability :  portable
--
-- Advanced features in databases operations 
--
-----------------------------------------------------------------------------

module Data.HDB.Advanced where

import Data.HListExtras
import Data.HDB.Representation
import Data.HDB.RelationalAlgebra
import Data.HDB.SQL
import Data.HDB.Databases
import Data.HDB.FunctionalDependencies
import Data.Map (Map)
import qualified Data.Map as Map
import qualified Data.Set as Set


-- * Transporting Functional Dependencies

-----------------------------------------------------------------------------------------------------------------
-- | Datatype for table with Functional Dependencies information

data TableWithFunDep  fds h k v => Table' h k v fds = Table' h (Map k v) fds deriving Show

class  HeaderFor h k v => TableWithFunDep fds h k v | fds h -> k v
instance ( 
        HeaderFor h k v,
        FDListFor fds (Table h k v)
      ) => TableWithFunDep  fds h k v

createTable' :: (
   TableWithFunDep fds h k v
 ) => h -> fds -> Table' h k v fds 
createTable' h fds = Table' h Map.empty fds

-- | selector for functional dependencies
tablefds (Table' _ _ fds) = fds

-- | Produce a table with fds from a normal table
addFDs (Table (a,b) m) = Table' (a,b) m ((FD a b) .*. HNil)

-----------------------------------------------------------------------------------------------------------------

-- | Computing set of FDList after projecting
projectValues' :: (
                  TableWithFunDep fds (a, b) k v,
                  TableWithFunDep fds' (a, b') k v',
                  LineUp b' v v' r,
                  ProjectFunDep b' fds fds'
                ) => b' -> Table' (a, b) k v fds -> Table' (a, b') k v' fds' 
projectValues' b' (Table' (a,b) m fds) = Table' (a,b') m' (projectFunDep b' fds)  
        where m' = Map.map (fst.lineUp b') m
-----------------------------------------------------------------------------------------------------------------
        

-- | Function that computes a new set of fds from a given set of fds and a 
 -- list of attributes : it keeps all fds that are valid only having attributes from b
class ProjectFunDep b fds fds' | b fds -> fds' where
    projectFunDep :: b -> fds -> fds'

instance ProjectFunDep b HNil HNil where
    projectFunDep _ _ = hNil
instance (
      FunDep x y,
      Difference x b x',
      HEq x' HNil bl, -- x only contains elements of b (new header)
      ProjectFD' bl b (HCons (FD x y) fds) fds'
    ) => ProjectFunDep b (HCons (FD x y) fds) fds' where
    projectFunDep b (HCons (FD x y) fds) = projectFD' bl b (HCons (FD x y) fds)
                where x' = difference x b
                      bl = hEq x' HNil

-- | Auxiliary function for projectFunDep
class ProjectFD' bl b fds fds' | bl b fds -> fds' where
    projectFD' :: bl -> b -> fds -> fds'


instance (FunDep x y, ProjectFunDep b fds fds')
        => ProjectFD' HFalse b (HCons (FD x y) fds) fds' where
    projectFD' _ b (HCons (FD x y) fds) = projectFunDep b fds
instance (
        FunDep x y,
        FunDep x y',
        Intersect b y y',
        ProjectFunDep b fds fds' 
      ) => ProjectFD' HTrue b (HCons (FD x y) fds) (HCons (FD x y') fds')  where
     projectFD' _ b (HCons (FD x y) fds) = HCons (FD x (intersect b y)) (projectFunDep b fds)
-----------------------------------------------------------------------------------------------------------------


-- | Remove a set of collumns from a table with fd information
removeColumns b' t@(Table' (a,b) m fds) = projectValues' (difference b b') t
-----------------------------------------------------------------------------------------------------------------

-- | Computing set of FDList after joining two tables

join' (Table' h@(a,b) m fds) (Table' h'@(a',b') m' fds') r = Table' h'' m'' fds''
        where Table h''@(a,ba'b') m'' = sqlInnerJoin (Table h m) (Table h' m') r
              newfd = FD a ba'b'
              fds'' = HCons newfd (union fds fds')
-----------------------------------------------------------------------------------------------------------------
  
-- | Checking insert table with FDS
--   can only insert a table if it respects at least the FDS of target table
insertTable' t'@(Table' h' mp' fds') 
             t@(Table' h mp fds) | (containAll h fds' fds) = Table' h m fds
                                 | otherwise = t
                  where Table h m = insertTable (Table h' mp') (Table h mp)
-----------------------------------------------------------------------------------------------------------------

-- | Check if a set of fds is contained in the closure of another set

class ContainAll h fds fds'  where 
    containAll :: h -> fds -> fds' -> Bool

instance ContainAll h HNil l where 
    containAll _ _ _ = True
instance (
           FunDep x y,
           FDList fds,
           Closure h x fds cl, 
           ContainedEq y cl b',
           BooleanValue b',
           ContainAll  h l fds 
         ) => ContainAll h (HCons (FD x y) l) fds  where
  containAll h (HCons (FD x y) l) fds = (booleanValue $ containedEq y cl) && (containAll h l fds)
                            where cl = closure h x fds

-----------------------------------------------------------------------------------------------------------------


-- * Database transformation

-- | Split a table according a fd

decompose (FD x y) (Table h@(a,b) m) = (Table (x,y) m', Table (a, b') m'')
    where b' = difference b y
          m' = projectMap (x,y) (a,b) m
          m'' = projectMap (a,b') (a,b) m

projectMap newh@(x,y) h@(a,b) m = m'
    where xy = hAppend x y
          ab = hAppend a b
          m' = Map.foldWithKey insrt Map.empty m
          insrt k v m = uncurry Map.insert (split (rtrv k v) (x,y)) m
          rtrv k v = lookupMany xy (Record $ hZip ab $ hAppend k v)
-----------------------------------------------------------------------------------------------------------------

-- | Produce a new database by replacing a table with two new tables
--   resulting from the decomposition according a given FD of the first table

decomposeByName rdb fd@(FD x y) tn tn1 tn2 = rdb'
    where 
        (t,fks) = rdb .!. tn
        (Table _ m,fks2) = rdb .!. tn1
        (Table _ m', t2@(Table (a,b) _)) = decompose fd t
        (fks1, _) = splitFKS tn tn1 tn2 (hAppend x y) (hAppend a b) fks
        t1 = Table (x,y) (Map.union m m')
        rdb' =  (tn1 .=. (t1,fks1)) .*. (tn2 .=. (t2,fks2) ) .*. (hDeleteAtLabel tn1 $ hDeleteAtLabel tn rdb)
-----------------------------------------------------------------------------------------------------------------

-- | Add new fks to a database from a set of fds that are known
--   to hold in the database

class AddFks rdb fds rdb' | rdb fds -> rdb' where
    addFks :: rdb -> fds -> rdb'

instance AddFks rdb HNil rdb where
    addFks rdb _ = rdb

instance (
    FD2FK rdb fd fk,
    AddFk rdb fk rdb'',
    AddFks rdb'' fds rdb'
    ) => AddFks rdb (HCons fd fds) rdb' where
    addFks rdb (HCons fd fds) = addFks (addFk rdb fk) fds
        where fk = fd2fk rdb fd

-- | Add a single fk to a relational database
--   The fk is added to the table that contains all atributtes in
--   the antecedent of the fk

class AddFk rdb fk rdb' | rdb fk -> rdb' where
    addFk :: rdb -> fk -> rdb'

instance AddFk (Record HNil) fk (Record HNil)
instance (
    HeaderFor (a,b) k v,
    HAppend a b ats,
    ContainedEq fk ats bl,
    AddFk' bl (Record (HCons (tn, (Table (a,b) k v,fks)) ts)) (FK fk tn' pk) rdb'
    ) => AddFk (Record (HCons (tn, (Table (a,b) k v,fks)) ts)) (FK fk tn' pk) rdb' where
    addFk rdb@(Record (HCons (tn, (Table (a,b) m,fks)) ts)) f@(fk, (tn',pk)) = addFk' bl rdb f
        where bl = containedEq fk (hAppend a b) 

-- | Auxiliar function to AddFk 
class AddFk' b rdb fk rdb' | b rdb fk -> rdb' where
    addFk' :: b -> rdb -> fk -> rdb'

instance 
    AddSingleFk fk fks fks'
    => AddFk' HTrue (Record (HCons (tn, (t, fks)) ts)) fk (Record (HCons (tn,(t, fks')) ts)) where
    addFk' _ (Record (HCons (tn, (t, fks)) ts)) fk =  Record $ HCons (tn,(t,addSingleFk fk fks)) ts
instance 
    AddFk (Record ts) fk (Record ts')
    => AddFk' HFalse (Record (HCons (tn, (t,fks)) ts)) fk (Record (HCons (tn, (t,fks)) ts')) where
    addFk' _ (Record (HCons (tn, (t,fks)) ts)) fk =  Record $ HCons (tn,(t,fks)) ts'
        where Record ts' = addFk (Record ts) fk

-- | Add a fk to a list of fks.
--   If the fk is already present in the list the same 
--   list is returned
class AddSingleFk fk fks fks' | fk fks -> fks' where
    addSingleFk :: fk -> fks -> fks'
instance (
    HMember fk fks b,
    AddSingleFk' b fk fks fks'
    ) => AddSingleFk fk (Record fks) (Record fks') where
    addSingleFk fk (Record fks) = Record $ addSingleFk' (hMember fk fks) fk fks

class AddSingleFk' b fk fks fks' | b fk fks -> fks' where
    addSingleFk' :: b -> fk -> fks -> fks'
instance AddSingleFk' HTrue fk fks fks where
    addSingleFk' _ _ fks = fks
instance AddSingleFk' HFalse fk fks (HCons fk fks) where
    addSingleFk' _ = HCons
-----------------------------------------------------------------------------------------------------------------

-- | Transform a fd in a fk 
--   Checks which table in the database has as primary key the consequent of the fd
--   and this table becomes the target table.

class FD2FK rdb fd fk | rdb fd -> fk where
    fd2fk :: rdb -> fd -> fk

instance (
    AttrList x, AttrList y, 
    PrimaryKey y rdb tn 
    ) => FD2FK rdb (FD x y) (FK x tn y) where
    fd2fk rdb (FD x y) = (x, (tn, y))
        where tn = primaryKey y rdb

-----------------------------------------------------------------------------------------------------------------
-- | Function that returns the table name in a database that has
--   a given set of attributes as primary key
class PrimaryKey y rdb tn | y rdb -> tn where
    primaryKey :: y -> rdb -> tn

instance (
    HeaderFor (a,b) k v,
    TypeEq y a bl,
    PrimaryKey' bl y tn (Record ts) tn'
    ) => PrimaryKey y (Record (HCons (tn,(Table (a,b) k v,fks)) ts)) tn' where 
    primaryKey y (Record (HCons (tn,(Table (a,b) m,fks)) ts)) = primaryKey' bl y tn (Record ts)
        where bl = typeEq y a

class PrimaryKey' b y tn rdb tn' | b y tn rdb -> tn' where
    primaryKey' :: b -> y -> tn -> rdb -> tn'

instance PrimaryKey' HTrue y tn rdb tn where 
    primaryKey' _ _ tn _  = tn

instance 
    PrimaryKey y rdb tn'
    => PrimaryKey' HFalse y tn rdb tn' where 
    primaryKey' _ y _ rdb  = primaryKey y rdb

-----------------------------------------------------------------------------------------------------------------
-- | Replace two tables in a database by a new table resulting 
--   of the composition of thos tables
--   Returns a pair with the new database and a set of fds (computed)
--   from all fks and fds that do not hold anymore in the new database
composeByName db fk tn1 tn2 newt = (rdb''', union (fd .*. fd' .*. HNil) (union r' r))
  where
    (_,fks1) = db .!. tn1
    (t2@(Table (a,b) _), fks2) = db .!. tn2
    (_, pk) = fks1 .!. fk
    on = (\r -> Record $ hZip pk $ lookupMany fk r)
    fks1'' = treatSelfRefs tn1 newt fks1'
    (fks2',r') = filterFks tn2 fks2 -- delete any self-reference for the 2nd table
    fks2'' = treatSelfRefs tn1 newt fks2'
    rdb' = hDeleteAtLabel tn2 db
    (rdb'',r) = mapfilterFks tn2 rdb' 
    (t1,fks1') = rdb'' .!. tn1
    fd = FD a b
    fd' = FD fk pk
    _ = DB db
    _ = DB rdb'''
    t3 = sqlInnerJoin t1 t2 on
    rdb''' = (newt .=. (t3, hLeftUnion fks1'' fks2'')) .*. (hDeleteAtLabel tn1 rdb'')

-- | Same as composeByName, but keeps the second table involved in the composition
--   Performs renaming in the attributes of this table.
composeByName' db fk tn1 tn2 newt = (rdb''', union (fd .*. fd' .*. HNil)  r)
  where
    (_,fks1) = db .!. tn1
    (t2@(Table (a,b) m2), fks2) = db .!. tn2
    (_, pk) = fks1 .!. fk
    on = (\r -> Record $ hZip pk $ lookupMany fk r)
    fks1'' = treatSelfRefs tn1 newt fks1'
    (fks2',r') = filterFks tn2 fks2 -- delete any self-reference for the 2nd table
    fks2'' = treatSelfRefs tn1 newt fks2'
    rdb' = hDeleteAtLabel tn2 db
    (rdb'',r) = mapfilterFks tn2 rdb' 
    (t1,fks1') = rdb'' .!. tn1
    fd = FD a b
    fd' = FD fk pk
    _ = DB db
    t3 = sqlInnerJoin t1 t2 on
    t2' = Table (a',b') m'
    a' = rename a newt
    b' = rename b newt
    fks2n = renamePartial (hAppend a b) fks2 newt -- rename fks 
    ResultList _ ks = select True fk t1 (const True)
    m' = mapDeleteMany ks m2
    rdb''' = (newt .=. (t3, hLeftUnion fks1'' fks2')) .*. (tn2 .=. t2' .=. fks2n) .*. (hDeleteAtLabel tn1 rdb'')

-- Delete a list of keys in a map

mapDeleteMany [] m = m
mapDeleteMany (k:ks) m = mapDeleteMany ks (Map.delete k m)

-----------------------------------------------------------------------------------------------------------------
-- | Function that performs renaming in a list of attributes, adding a prefix 
class Rename ats newID ats' | ats newID -> ats' where
    rename :: ats -> newID -> ats'

instance Rename HNil id HNil where
    rename _ _ = hNil

instance Rename ats newID ats' 
    => Rename (HCons (Attribute t nr) ats) newID (HCons (Attribute t (nr,newID)) ats') where
    rename (HCons at ats) newID = HCons (undefined :: (Attribute t (nr,newID))) (rename ats newID)

instance Rename ats newID ats' 
    => Rename (HCons (DefaultableAttribute t nr) ats) newID (HCons (DefaultableAttribute t (nr,newID)) ats') where
    rename (HCons (Default v) ats) newID = HCons ((Default v):: (DefaultableAttribute t (nr,newID))) (rename ats newID)

instance Rename ats newID ats' 
    => Rename (HCons (NullableAttribute t nr) ats) newID (HCons (NullableAttribute t (nr,newID)) ats') where
    rename (HCons at ats) newID = HCons (undefined :: (NullableAttribute t (nr,newID))) (rename ats newID)

-- | Function that performs renaming in a list of attributes, adding a prefix
--   Only attributes in 'restriction' are renamed
class RenamePartial restriction ats newID ats' | restriction ats newID -> ats' where
    renamePartial :: restriction -> ats -> newID -> ats'

instance RenamePartial rest HNil id HNil where
    renamePartial _ _ _ = hNil

instance (
    RenamePartial rest ats newID ats',
    HMember at rest bl,
    RenamePartial' bl at newID at'
    ) => RenamePartial rest (HCons at ats) newID (HCons at' ats') where
    renamePartial rest (HCons at ats) newID = HCons at' (renamePartial rest ats newID)
        where at' = renamePartial' (hMember at rest) at newID 
              

instance (
    RenamePartial rest fk newID fk',
    RenamePartial rest pk newID pk',
    RenamePartial rest fks newID fks' 
    ) => RenamePartial rest (HCons (FK fk tn pk) fks) newID (HCons (FK fk' tn pk') fks') where
    renamePartial rest (HCons (fk,(tn,pk)) fks) newID = HCons (fk', (tn,pk')) (renamePartial rest fks newID)
        where --at' = renamePartial' (hMember at rest) at newID 
              fk' = renamePartial rest fk newID 
              pk' = renamePartial rest pk newID 

instance RenamePartial rest ats newID ats' 
    => RenamePartial rest (Record ats) newID (Record ats') where
    renamePartial rest (Record ats) newID = Record $ renamePartial rest ats newID

class RenamePartial' bl at newID at' | bl at newID -> at' where
    renamePartial' :: bl -> at -> newID -> at'

instance RenamePartial' HFalse at newID at where
    renamePartial' _ at _ = at

instance RenamePartial' HTrue (Attribute t nr) newID (Attribute t (nr,newID)) where
    renamePartial' _ at newID = undefined :: Attribute t (nr,newID)
instance RenamePartial' HTrue (DefaultableAttribute t nr) newID (DefaultableAttribute t (nr,newID)) where
    renamePartial' _ (Default v) newID = (Default v) :: DefaultableAttribute t (nr,newID)
instance RenamePartial' HTrue (NullableAttribute t nr) newID (NullableAttribute t (nr,newID)) where
    renamePartial' _ at newID = undefined :: NullableAttribute t (nr,newID)
-----------------------------------------------------------------------------------------------------------------


-- | Class that puts all fks in a list that refer to a certain tablename as self references to the new table
class TreatSelfRefs tn newt fks fks' | tn newt fks -> fks' where
    treatSelfRefs :: tn -> newt -> fks -> fks'

instance TreatSelfRefs tn newt (Record HNil) (Record HNil) where
    treatSelfRefs _ _ _ = Record HNil

instance (
    TypeEq tn tn' b,
    TreatSelfRefs' b tn newt (Record (HCons (fk, (tn',pk)) fks)) fks' 
    ) => TreatSelfRefs tn newt (Record (HCons (fk, (tn',pk)) fks)) fks' where
    treatSelfRefs tn newt fs@(Record (HCons (fk, (tn',pk)) fks)) = 
        treatSelfRefs' (typeEq tn tn') tn newt fs

class TreatSelfRefs' b tn newt fks fks' | b tn newt fks -> fks' where
    treatSelfRefs' :: b -> tn -> newt -> fks -> fks'

instance ( 
    AttrList fk, AttrList pk,
    TreatSelfRefs tn newt (Record fks) (Record fks')
    ) => TreatSelfRefs' HTrue tn newt (Record (HCons (fk, (tn', pk)) fks)) (Record (HCons (FK fk newt pk) fks')) where
    treatSelfRefs' _ tn  newt (Record (HCons (fk, (tn', pk)) fks)) = Record (HCons (fk, (newt, pk)) fks')
        where Record fks' = treatSelfRefs tn newt (Record fks)

instance 
    TreatSelfRefs tn newt (Record fks) (Record fks')
     => TreatSelfRefs' HFalse tn newt (Record (HCons fk fks)) (Record (HCons fk fks')) where
    treatSelfRefs' _ tn newt (Record (HCons fk fks)) = Record (HCons fk fks')
        where Record fks' = treatSelfRefs tn newt (Record fks)

--------------------------------------------------------------------------------------


-- | Class that eliminates all fks in a list that refer to a certain tablename
class FilterFks tn fks fks' r | tn fks -> fks' r where
    filterFks :: tn -> fks -> (fks',r)

instance FilterFks tn (Record HNil) (Record HNil) HNil where
    filterFks _ _ = (Record HNil, HNil)

instance (
    TypeEq tn tn' b,
    FilterFks' b tn (Record (HCons (fk, (tn',pk)) fks)) fks' r'
    ) => FilterFks tn (Record (HCons (fk, (tn',pk)) fks)) fks' r' where
    filterFks tn fs@(Record (HCons (fk, (tn',pk)) fks)) = 
        filterFks' (typeEq tn tn') tn fs

class FilterFks' b tn fks fks' r' | b tn fks -> fks' r' where
    filterFks' :: b -> tn -> fks -> (fks', r')

instance ( 
    AttrList fk, AttrList pk,
    FilterFks tn (Record fks) fks' r
    ) => FilterFks' HTrue tn (Record (HCons (fk, (tn',pk)) fks)) fks' (HCons (FD fk pk) r) where
    filterFks' _ tn  (Record (HCons (fk, (tn',pk)) fks)) = (fks', HCons fd r)
        where (fks',r) = filterFks tn (Record fks)
              fd = FD fk pk
instance 
    FilterFks tn (Record fks) (Record fks') r
     => FilterFks' HFalse tn (Record (HCons fk fks)) (Record (HCons fk fks')) r where
    filterFks' _ tn (Record (HCons fk fks)) = (Record (HCons fk fks'), r)
        where (Record fks', r) = filterFks tn (Record fks)

-- | Class that eliminates all fks in a database that refer to a certain tablename
class MapFilterFks tn rdb rdb' r | tn rdb -> rdb' r where
    mapfilterFks :: tn -> rdb -> (rdb',r)

instance MapFilterFks tn (Record HNil) (Record HNil) HNil where
    mapfilterFks _ _ = (Record HNil, HNil)

instance (
    FilterFks tn fks fks' r',
    MapFilterFks tn (Record ts) (Record ts') r'',
    HAppend r' r'' r
    ) => MapFilterFks tn (Record (HCons (tn', (t, fks)) ts)) (Record (HCons (tn', (t,fks')) ts')) r where
    mapfilterFks tn (Record (HCons (tn', (t, fks)) ts)) = (Record (HCons (tn', (t,fks')) ts'),hAppend r' r'')
        where (fks', r') = filterFks tn fks
              (Record ts', r'') = mapfilterFks tn (Record ts) 

--------------------------------------------------------------------------------------

class SplitFKS tn1 tn2 tn3 h h' fks fks1 fks2 | tn1 tn2 tn3 h h' fks -> fks1 fks2 where
    splitFKS :: tn1 -> tn2 -> tn3 -> h -> h' -> fks -> (fks1,fks2)

instance SplitFKS tn1 tn2 tn3 h h' (Record HNil) (Record HNil) (Record HNil) where
    splitFKS _ _ _ _ _ _ = (Record HNil,Record HNil)

instance (
    ContainedEq fk h b,
    ContainedEq fk h' b',
    TypeEq tn1 t b'',
    SplitFKS' b b' b'' tn1 tn2 tn3 h h' (Record (HCons (fk, (t, pk)) fks)) fks1 fks2
    ) => SplitFKS tn1 tn2 tn3 h h' (Record (HCons (fk, (t, pk)) fks)) fks1 fks2 where
    splitFKS tn1 tn2 tn3 h h' fs@(Record (HCons (fk, (t, pk)) fks)) = splitFKS' (containedEq fk h) (containedEq fk h') (typeEq tn1 t) tn1 tn2 tn3 h h' fs

class SplitFKS' b b' b'' tn1 tn2 tn3 h h' fks fks1 fks2 | b b' b'' tn1 tn2 tn3 h h' fks -> fks1 fks2 where
    splitFKS' :: b -> b' -> b'' -> tn1 -> tn2 -> tn3 -> h -> h' -> fks -> (fks1,fks2)

-- self reference -- the first table will keep it
instance SplitFKS tn1 tn2 tn3 h h' (Record fks) (Record fks1) fks2 
    => SplitFKS' HTrue HFalse HTrue tn1 tn2 tn3 h h' (Record (HCons (FK fk t pk) fks)) (Record (HCons (FK fk tn3 pk) fks1)) fks2 where
    splitFKS' _ _ _ tn1 tn2 tn3 h h' (Record (HCons (fk, (t, pk)) fks)) = (Record $ HCons (fk, (tn3, pk)) fks1,fks2)
            where (Record fks1,fks2) = splitFKS tn1 tn2 tn3 h h' (Record fks)


-- no self reference 
instance SplitFKS tn1 tn2 tn3 h h' (Record fks) (Record fks1) fks2 
    => SplitFKS' HTrue HFalse HFalse tn1 tn2 tn3 h h' (Record (HCons fk fks)) (Record (HCons fk fks1)) fks2 where
    splitFKS' _ _ _ tn1 tn2 tn3 h h' (Record (HCons fk fks)) = (Record $ HCons fk fks1,fks2)
            where (Record fks1,fks2) = splitFKS tn1 tn2 tn3 h h' (Record fks)

-- self reference -- the second table will keep it
instance SplitFKS tn1 tn2 tn3 h h' (Record fks) fks1 (Record fks2)
    => SplitFKS' HFalse HTrue HTrue tn1 tn2 tn3 h h' (Record (HCons  (FK fk t pk) fks)) fks1 (Record (HCons  (FK fk tn2 pk) fks2)) where
    splitFKS' _ _ _ tn1 tn2 tn3 h h' (Record (HCons (fk, (t, pk)) fks)) = (fks1,Record $ HCons (fk, (tn2, pk)) fks2)
            where (fks1,Record fks2) = splitFKS tn1 tn2 tn3 h h' (Record fks)

-- no self reference -- the second table keeps the fk
instance SplitFKS tn1 tn2 tn3 h h' (Record fks) fks1 (Record fks2)
    => SplitFKS' HFalse HTrue HFalse tn1 tn2 tn3 h h' (Record (HCons fk fks)) fks1 (Record (HCons fk fks2)) where
    splitFKS' _ _ _ tn1 tn2 tn3 h h' (Record (HCons fk fks)) = (fks1,Record $ HCons fk fks2)
            where (fks1,Record fks2) = splitFKS tn1 tn2 tn3 h h' (Record fks)
-----------------------------------------------------------------------------------------------------------------

-- * On the level of individual tables.

-- | Add a column.
addColumn :: (
  HeaderFor (a,b) k v,
  HeaderFor (a, HCons c b) k (HCons d v)
 ) => c -> (k -> v -> d) -> Table (a,b) k v
   -> Table (a, HCons c b) k (HCons d v)
addColumn c f (Table (a,b) mp)
  = Table (a,HCons c b) (Map.mapWithKey (\k v -> HCons (f k v) v) mp)

-- | Remove a column.
rmColumn :: (
  HeaderFor (a,b) k v,
  HFind c b n,
  HDeleteAtHNat n b b',
  HDeleteAtHNat n v v',
  HeaderFor (a,b') k v'
 ) => c -> Table (a,b) k v -> Table (a,b') k v'
rmColumn c (Table (a,b) mp)
 = Table (a,b') mp'
   where
     n = hFind c b
     b' = hDeleteAtHNat n b
     mp' = Map.map (hDeleteAtHNat n) mp



-----------------------------------------------------------------------------
-- * Operations lifted to the database level

insertInto :: (
  HZip a v' row,
  ModifyAtLabel tn (Table h k v,fk) (Table h k v,fk) rdb rdb,
  LineUp h (Record row) (k,v) (Record HNil),
  HeaderFor h k v
 ) => rdb -> tn -> a -> v' -> rdb
insertInto rdb tn as vs = modifyAtLabel tn insrt rdb
 where
   insrt (t,fk) = (insertValues' t as vs,fk)

selectInto rdb d a tns w o tn = modifyAtLabel tn f rdb
 where
   ts = fst $ hUnzip $ lookupMany tns rdb
   f (t,fk) = (insertResultList a (select''' d a ts w o) t,fk)


createTable (Record ts) tn h = Record (tn .=. (t,fks) .*. ts)
  where
    t = Table h Map.empty
    fks = Record HNil

alterTableAddColumn rdb tn c = modifyAtLabel tn f rdb
  where
    f (Table (a,b) m,fk) = (Table (a, hAppend b c) (Map.map add m),fk)
    add v = hAppend v (snd $ lineUpPadd c HNil)
-----------------------------------------------------------------------------------------------------------------

-- * Mapper function

-- a - target attributes
-- b - domain attributes
data MapperFunction a b x y = MapperFunction a b (x->[y])

mapperM f m =
  Map.foldWithKey (\k v l -> (f (hAppend k v))++l) [] m

class MapperT t m l | t m -> l where
    mapperT :: t -> m -> l
instance (
    RESULTSET b kv',
    HeaderFor h k v,
    LookupMany b r kv',
    Row h k v r
    ) => MapperT (Table h k v) (MapperFunction a b kv' y) [y] where
    mapperT t (MapperFunction tg s f) =  concat $ map f $ Set.toList st
                where ResultSet _ st = project s t

class Mapper t mps rls | t mps -> rls where
    mapper' :: t -> mps -> rls

instance Mapper t HNil HNil where
    mapper' _ _ = HNil
instance ( 
    Mapper (Table (ka,va) k v) mps rls,
    HeaderFor (ka,va) k v,
    AttributesFor a y,
    MapperT (Table (ka,va) k v) (MapperFunction a b x y) [y]
    ) => Mapper (Table (ka,va) k v) (HCons (MapperFunction a b x y) mps) (HCons (ResultList a y) rls) where
    mapper' t (HCons mp@(MapperFunction a _ _) mps) = HCons (ResultList a (mapperT t mp)) (mapper' t mps)


mapper t lmps = products $ mapper' t lmps

-----------------------------------------------------------------------------------------------------------------

-- Example of mapper
acctId = undefined :: Attribute String ()
acctId' = undefined :: Attribute String Zero
am = undefined :: Attribute Int One
amount = undefined :: Attribute Int Two
seqno = undefined :: Attribute Int Three

acct = MapperFunction (acctId'.*.HNil) (acctId.*.HNil) (padd 4)
amt = MapperFunction (amount.*.seqno.*.HNil) (am.*.HNil) amtf

amtf (HCons am HNil) | ((am `mod` 100) /= 0) =  (amtf $ HCons (am - am`mod`100) HNil) ++ [(am`mod`100) .*. ((am `div` 100)+1) .*. HNil]
                     | otherwise = map pair2hlist $ zip (replicate (am `div` 100) 100) [1..]

loans = Table (acctId.*.HNil, am .*. HNil) $
        Map.insert ("12".*.HNil) (10.*.HNil) $
        Map.insert  ("3456".*.HNil) (140.*.HNil) $
        Map.insert  ("901".*.HNil) (250.*.HNil) $
        Map.empty

payments = ResultList (acctId' .*. amount .*. seqno .*. HNil) []

padd i (HCons s HNil) = let j = i - length s 
                            s' = replicate j '0'
                        in [HCons (s' ++ s) HNil]

class Pair2HList p l | p -> l where
    pair2hlist :: p -> l

instance Pair2HList (a,b) (HCons a (HCons b HNil)) where
    pair2hlist (a,b) = (HCons a (HCons b HNil))


-----------------------------------------------------------------------------------------------------------------
