{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances -fallow-undecidable-instances #-}
-----------------------------------------------------------------------------
-- Maintainer  :  Alexandra and Joost Visser
-- Stability   :  experimental
-- Portability :  portable
--
-- Advanced features in databases operations -- Normalization, lossless decomposition
--
-----------------------------------------------------------------------------

module Data.HDB.FunctionalDependencies where

import Data.HListExtras
import Data.HDB.Representation
import Data.HDB.RelationalAlgebra
import Data.HDB.SQL()
import qualified Data.Map as Map
import qualified Data.Set as Set
-----------------------------------------------------------------------------
-- * Functional Dependencies

-- | constraint for functional dependency : both antecedent (x)
    -- and consequent (y) must be list of attributes
class FunDep x y 
instance (AttrList at, AttrList at') 
        => FunDep at at'

-- | datatype for Functional Dependency
data FunDep x y => FD x y = FD x y deriving Show

-- | List of functional dependencies
class FDList at
instance FDList HNil
instance (FunDep lhs rhs, FDList fds) 
        => FDList (HCons (FD lhs rhs) fds)

-- | Checking that a list only has attributes
class AttrList ats
instance AttrList HNil
instance (AttributeFor at t, AttrList l) => AttrList (HCons at l)

-- | Checking that a list of FD's is valid for a table, meaning that all 
   -- attributes used in FD are in the table header
class FDListFor fds t 
instance (
        AttrListFor fds ats,
        HAppend a b h,
        HMember (FD a b) fds HTrue,
        ContainedEq ats h HTrue
        ) => FDListFor fds (Table (a,b) ks vs)

-- | computing the list of attributes used in a list of FD's
class AttrListFor fds ats | fds -> ats 
instance AttrListFor HNil HNil
instance (
      Union x y z,
      AttrListFor fds ats',
      Union z ats' ats
    ) => AttrListFor (HCons (FD x y) fds) ats

atListFor :: AttrListFor fds ats => fds -> ats
atListFor = undefined

-- | Checking if an attribute is a super key
class IsSuperKey s all fds b | s all fds -> b
instance (
      PowerSet s ss,
      FilterEmptySet ss ss',
      MapIsKey ss' all fds b 
    ) => IsSuperKey s all fds b

-- | Checking if all attributes in a list are keys
-- equivalent to: or [isKey x | x <- listAttributes]

class MapIsKey ss all fds b | ss all fds -> b
instance MapIsKey HNil all fds HFalse
instance (
      IsKey x all fds b',
      MapIsKey xs all fds b'',
      HOr b' b'' b
    ) => MapIsKey (HCons x xs) all fds b

mapIsKey :: MapIsKey ats all fds b => ats -> all -> fds -> b
mapIsKey = undefined


-- Prime Attributte : A is a prime attributte if A is member of any key
-- | Checking if an attribute is prime
class IsPrime at all fds b | at all fds -> b
instance (
   Keys all fds lk,
   MapHMember at lk b
 ) =>
    IsPrime at all fds b 

isPrime :: IsPrime at all fds b => at -> all -> fds -> b
isPrime = undefined

-- | Keys -- Deriving list of Keys based on FDList 

class Keys all fds listOfKeys | all fds -> listOfKeys 

instance (
 ProperSubsets all ps,
 FilterKeys ps all fds listOfKeys 
 ) => Keys all fds listOfKeys

-- test purposes only
keys :: (Keys all fds listOfKeys) => all -> fds -> listOfKeys
keys = undefined

-- | Filter keys from a list of ALL possible combination of attributes -- that's inefficient...
class FilterKeys ps all fds lKeys | ps all fds -> lKeys  
instance FilterKeys HNil all fds HNil
instance (
   IsKey s all fds b,
   FilterKeys' b (HCons s l) all fds lkeys
 ) => FilterKeys (HCons s l) all fds lkeys



fkeys :: FilterKeys ps all fds lks => ps -> all -> fds -> lks
fkeys = undefined

-- Auxiliary to FilterKeys
class FilterKeys' b ps all fds lKeys | b ps all fds -> lKeys  

instance  FilterKeys l all fds rkeys => FilterKeys' HTrue (HCons k l) all fds (HCons k rkeys)
instance  FilterKeys l all fds rkeys => FilterKeys' HFalse (HCons k l) all fds rkeys

-- | Predicate that checks if an attributte list is a key
-- isKey -- page 383 UllMan

class IsKey x all fds b | x all fds -> b
instance (
  Closure all x fds cl, -- this 2 conditions guarantee  that X -> All in in X+
  ContainedEq all cl b', 
  -- now we must gurantee that for no proper subset Y of X,  Y -> All is in F+
  Minimal x all fds b'',
  HAnd b' b'' b
 ) => IsKey x all fds b

isKey :: IsKey x all fds b => x -> all -> fds -> b
isKey = undefined
isSKey :: IsSuperKey x all fds b => x -> all -> fds -> b
isSKey = undefined


-- | Checking if a candidate key is minimal
class Minimal x all fds b | x all fds -> b
instance  (
     PowerSet x xs,
     FilterEmptyTotal x xs xs', -- getting only the proper subsets
     MapIsNotInFDClosure xs' all fds b
    ) => Minimal x all fds b 

-- just for test purposes
minimal :: Minimal x all fds b => x -> all -> fds -> b
minimal = undefined

-- | Checking if a set of Attributes is in FD+ (closure of set of FDS)
class IsInFDClosure y x fds b | y x fds -> b
instance (
   Closure y x fds cl, -- this 2 conditions guarantee  that X -> Y in in X+
   ContainedEq y cl b
 ) => IsInFDClosure y x fds b

-- just for test purposes
isInFDClosure :: IsInFDClosure all x fds b => all -> x -> fds -> b
isInFDClosure = undefined


-- | Checking if all sets of Attributes contained in a list are NOT in FD+ (closure of set of FDS)
class MapIsNotInFDClosure xs all fds b | xs all fds -> b
instance MapIsNotInFDClosure HNil all fds HTrue
instance (
 IsInFDClosure all x fds b1,
 Not b1 b',
 MapIsNotInFDClosure xs all fds b'',
 HAnd b' b'' b)
 => MapIsNotInFDClosure (HCons x xs) all fds b


-- | Computation of the closure of a Set of Attributtes with respect to a Set of FD's
-- Algorithm 7.1, page 388, Ullman "Principles  on Databases and Knowledge Base Systems"


class Closure u x fd closure | u x fd -> closure where
    closure ::  u -> x -> fd -> closure 

-- Dummy instance
instance Closure u x HNil HNil

instance 
    (
     TypeEq u x b, 
     Closure' b u x fd cl
    )
    => Closure u x fd cl where
    closure u x fd = closure' (typeEq u x) u x fd

class Closure' b u x fd closure | b u x fd -> closure where
    closure' ::  b -> u -> x -> fd -> closure 

instance Closure' HTrue u x fd x where
    closure'  _ _ x _ = x

instance 
    (GetAntecedents x fd rhs,
     Union x rhs x' ,
     Closure x x' fd closure
    ) => Closure' HFalse u x fd closure where
  closure' _ u x fd = closure x (x `union` (getAntecedents x fd)) fd 

class GetAntecedents x l l' | x l -> l' where
    getAntecedents :: x -> l -> l' 

instance GetAntecedents x HNil HNil where
    getAntecedents _ _ = hNil

instance (
      FunDep lhs rhs,
   Contained lhs x b,
   GetAntecedents' b x (HCons (FD lhs rhs) fds) c 
   )
   => GetAntecedents x (HCons (FD lhs rhs) fds) c where
    getAntecedents x (HCons (FD lhs rhs) fds)  = getAntecedents' (contained lhs x) x (HCons (FD lhs rhs) fds) 

class GetAntecedents' b x l l' | b x l -> l' where
    getAntecedents' :: b -> x -> l -> l' 
    
instance (
      FunDep lhs rhs,
   GetAntecedents x fds cl',
   Union rhs cl' cl
 ) => GetAntecedents' HTrue x (HCons (FD lhs rhs) fds) cl where
    getAntecedents' _ x (HCons (FD lhs rhs) fds) = union rhs (getAntecedents x fds)

instance (
      FunDep lhs rhs,
   GetAntecedents x fds cl
   )
   => GetAntecedents' HFalse x  (HCons (FD lhs rhs) fds) cl where
    getAntecedents' _ x (HCons (FD lhs rhs) fds) = getAntecedents x fds


-- | Fully dependent definition (Maier page 99)

class FullyDep fds x y b |  fds x y -> b 

instance LeftReduced fds x y b => FullyDep fds x y b

class LeftReduced fds x y b | fds x y -> b 

instance ( FunDep x y,
           ProperSubsets x xs,
           LeftReduced' xs y fds b
          ) => LeftReduced fds x y b

class LeftReduced' xs y fds b | xs y fds -> b 

instance LeftReduced' HNil y fds HTrue

instance ( FunDep x y,
           IsInFDClosure y x fds b1,
           Not b1 b',
           LeftReduced' xs y fds b'',
           HAnd b' b'' b
          ) => LeftReduced' (HCons x xs) y fds b


-- | just for test purposes
leftr :: (LeftReduced fds x y b) => fds -> x -> y -> b
leftr = undefined

fullydep :: (FullyDep fds x y b) => fds -> x -> y -> b
fullydep = undefined

-- | Check if a single FD holds in a given table
-- (class just used to gather the constraints)

class CheckFD fd t where 
    checkFD :: fd -> t -> Bool

instance (  HeaderFor h k v,
            RESULTSET xy kv',
            LookupMany xy r kv', 
            Row h k v r,
            HZip xy kv' r,
            LookupMany x (Record r) k',
            Ord k',
            HAppend x y xy,
            FunDep x y
         ) => CheckFD (FD x y) (Table h k v) where 
    checkFD (FD x y) t = Map.fold worker True (groupBy x $ project xy t)
                where worker set b =  b && (Set.size set == 1)
                      xy = hAppend x y

-- | Check if a list of FD's holds in a table
class CheckFDs fds t where
    checkFDs :: fds -> t -> Bool

instance CheckFDs HNil t where
    checkFDs HNil t = True

instance (
    CheckFD (FD x y) (Table h k v), 
    CheckFDs fds (Table h k v)
  ) => CheckFDs (HCons (FD x y) fds) (Table h k v) where
    checkFDs (HCons fd fds) t = (checkFD fd t) && checkFDs fds t 


