{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances -fallow-undecidable-instances -fallow-incoherent-instances #-}
-----------------------------------------------------------------------------
-- Maintainer  :  Alexandra and Joost Visser
-- Stability   :  experimental
-- Portability :  portable
--
-- Advanced features in databases operations -- Normalization
--
-----------------------------------------------------------------------------

module Data.HDB.NormalForms where

import Data.HListExtras
import Data.HDB.FunctionalDependencies
import Data.HDB.Representation
import Data.Set (Set)
import Data.Map (Map)


-- * Normal Forms

-- | Checking if a list of attributes is in 1stNF
class Is1stNF l b | l -> b
instance Is1stNF HNil HTrue
instance (Atomic at b, Is1stNF ats b', HAnd b b' b'') => Is1stNF (HCons at ats) b''

class Atomic at b | at -> b
instance Atomic t b => Atomic (Attribute t nr) b
instance ( IsList t b1, 
           IsSet t b2, 
           IsMap t b3, 
           HOr b1 b2 b'', 
           HOr b3 b'' b', 
           Not b' b
         ) =>  Atomic t b


class HBool b => IsList t b | t -> b

instance (HBool b', TypeEq t Char b, Not b b')=> IsList [t] b' -- excludes Strings
instance CFalse t b => IsList t b

class HBool b => CFalse t b | t -> b
instance TypeEq t [t] b => CFalse t b

class IsSet t b | t -> b
instance IsSet (Set t) HTrue
instance CFalse t b => IsSet t b
 
class IsMap t b | t -> b
instance IsMap (Map k a) HTrue
instance CFalse t b => IsMap t b
 

is1stNF :: Is1stNF l b => l -> b 
is1stNF = undefined

-- | 
{-
Definition for 2nd NF 
A relation scheme is in 2ndNF if it is in 1stNF and every non-prime attributte is fully dependent
on every key of R 
-}

class Is2ndNF all fds b | all fds -> b
instance ( Is1stNF all HTrue,
           NonPrimes all lk nonprimes,
           Keys all fds lk,
           MapFullyDep fds lk nonprimes b
         )
          => Is2ndNF all fds b

class MapFullyDep fds atlist atlist' b | fds atlist atlist' -> b

instance  MapFullyDep fds HNil atlist' HTrue
instance (
           MapFullyDep' fds k ats b1,
           MapFullyDep fds ks ats b2 ,
           HAnd b1 b2 b 
         ) => MapFullyDep fds (HCons k ks) ats b

class MapFullyDep' fds atlist atlist' b | fds atlist atlist' -> b

instance  MapFullyDep' fds k HNil HTrue

instance (
           FullyDep fds k (HCons at HNil) b1,
           MapFullyDep' fds k ats b2 ,
           HAnd b1 b2 b 
         ) => MapFullyDep' fds k  (HCons at ats) b

class NonPrimes all lk nonprimes | all lk -> nonprimes  
instance (Unions lk lks, Difference all lks np) => NonPrimes all lk np

nonprimes :: (NonPrimes all lk np) => all -> lk -> np
nonprimes = undefined

mapfd :: (MapFullyDep fds ks ats b) => fds -> ks -> ats -> b
mapfd = undefined

is2ndNF :: (Is2ndNF all fds b) => all -> fds -> b
is2ndNF = undefined


{-
Ullman page 402
A relation Schema R (header) is in 3rd NF if whenever X -> A (A single Attributte)
holds in R and A is not in X, then either X is a superkey for R or A is Prime

X is a superKey if there exists X' subsetEq X such that X is a key
-}

class Is3rdNF h fds b | h fds -> b
instance Is3rdNF h HNil HTrue
instance (Is1stNF h HTrue, Is3rdNF' h (HCons e l) (HCons e l) b) => Is3rdNF h (HCons e l) b


is3rdNF :: Is3rdNF h l b => h -> l -> b 
is3rdNF = undefined

class Is3rdNF' h fds allfds b | h fds allfds -> b
instance Is3rdNF' h HNil allfds HTrue
instance (
    HMember y x bb, -- y is a single attributte
    Not bb bYnotinX,
    Is3rdNFAtomic bYnotinX h x y fds b',
    Is3rdNF' h fds' fds b'',
    HAnd b' b'' b
  ) => Is3rdNF' h (HCons (FD x (HCons y HNil)) fds') fds b

-- | check if the head of fd's list does not violate 3rd NF
class Is3rdNFAtomic check h x y fds b | check h x y fds -> b
instance Is3rdNFAtomic HFalse h x y fds HTrue -- A is in X 
instance (
    IsSuperKey x h fds issuperkey,
    IsPrime y h fds isprime,
    HOr issuperkey isprime b
  ) => Is3rdNFAtomic HTrue h x y fds b


is3rdNFAt :: Is3rdNFAtomic b' h x y l b => b' -> h -> x -> y -> l -> b 
is3rdNFAt = undefined

{-
Boyce-Codd NF
Ullman page 401
A relation Schema R (header) is in BoyceCodd NF if whenever X -> A holds in 
R an A is not in X, then X is a superkey for R  MORE RESTRICTIVE than 3rdNF
-}

class BoyceCoddNF h fds b | h fds -> b
instance BoyceCoddNF h HNil HTrue
instance (Is1stNF h HTrue, BoyceCoddNF' h (HCons e l) (HCons e l) b) => BoyceCoddNF h (HCons e l) b

boyceCoddNF :: BoyceCoddNF h l b => h -> l -> b 
boyceCoddNF = undefined


class BoyceCoddNF' h fds allfds b | h fds allfds -> b
instance BoyceCoddNF' h HNil fds HTrue
instance (
    HMember y x bb, -- y is a single attributte
    Not bb bYnotinX,
    BoyceCoddNFAtomic bYnotinX h x fds b',
    BoyceCoddNF' h fds' fds b'',
    HAnd b' b'' b
  ) => BoyceCoddNF' h (HCons (FD x (HCons y HNil)) fds') fds b

class BoyceCoddNFAtomic check h x fds b | check h x fds -> b
instance BoyceCoddNFAtomic HFalse h x fds HTrue -- A is in X 
instance (
    IsSuperKey x h fds b 
  ) => BoyceCoddNFAtomic HTrue h x fds b


