-----------------------------------------------------------------------------
-- |
-- Maintainer  :  Alexandra Silva, Joost Visser
-- Stability   :  experimental
-- Portability :  portable
--
-- Representation of basic elements needed to construct a relational database.
--
-----------------------------------------------------------------------------

module Data.HDB.Representation where

import qualified Data.Map as Map
import Data.Map (Map)
import qualified Data.Set as Set
import Data.Set (Set)
import Data.HListExtras
import Prelude hiding (null)

-----------------------------------------------------------------------------
-- * Representation of attributes

-- | Attributes take two phantom arguments.
--   One indicates the column type.
--   The other can be used to disambiguate 
--   when several attributes have the same column type.
--   There are three types of attributes : normal attributes,
--   which are non-nullables but can have system default 
--   values, nullable attributes and user-declared 
--   defaultable attributes
data Attribute t name 
data NullableAttribute t name
data DefaultableAttribute t name = Default t

attribute = undefined :: Attribute t name
nullableattribute = undefined :: NullableAttribute t name

mkDummy :: AttributeFor at v => at -> v
mkDummy _ = undefined

-- | Instances for Eq and Show 
-- Show instances should be defined for each unique label
instance Eq (Attribute t name) where
    (==) a a' = True
instance Show (Attribute t name) where
    showsPrec _ a = shows "attribute"
instance Eq (NullableAttribute t name) where
    (==) a a' = True
instance Show (NullableAttribute t name) where
    showsPrec _ a = shows "nullable_attribute"
instance Eq t => Eq (DefaultableAttribute t name) where
    (==) (Default t) (Default t') = t==t'
instance Show (DefaultableAttribute t name) where
    showsPrec _ a = shows "defaultable_attribute"

-- | Function that derives the type of an attribute 
class AttributeFor a t | a -> t 
instance AttributeFor (Attribute t nr) t 
instance AttributeFor (DefaultableAttribute t nr) t 
instance AttributeFor (NullableAttribute t nr) (Maybe t) 

-- | A list of column types can be derived from an attribute list.
class AttributesFor a v  | a -> v
instance AttributesFor HNil HNil
instance (AttributeFor at t, AttributesFor a v)
   => AttributesFor (HCons at a) (HCons t v)

-- | A list of key column types can be derived from an attribute list, 
--   that cannot be nullable or defaultable
class KeyAttributesFor a v  | a -> v
instance KeyAttributesFor HNil HNil
instance KeyAttributesFor a v
   => KeyAttributesFor (HCons (Attribute t nr) a) (HCons t v)


-----------------------------------------------------------------------------
-- * Representation of tables

-- | Tables take three arguments: 
--   header, key type, and value type.
data HeaderFor h ks vs => Table h ks vs = Table h (Map ks vs) 
                                            deriving (Eq,Show,Read)

-- | Retrieve the tuples from a table.
tuples :: HeaderFor h ks vs => Table h ks vs -> Map ks vs
tuples (Table h m) = m

-- | Retrieve the header from a table.
header :: HeaderFor h ks vs => Table h ks vs -> h
header (Table h m) = h

-- | Compute the key and non-key types from a given table header.
class Ord kts
   => HeaderFor h kts nkts | h -> kts nkts
instance (
   HAppend kas nkas as, 
   NoRepeats as,
   KeyAttributesFor kas kts,
   AttributesFor nkas nkts,
   Ord kts
 ) => HeaderFor (kas,nkas) kts nkts
 
-- | A well-formed table header is a pair of attribute lists
--   where no attribute is repeated.
class TABLEHEADER h
instance (
   HAppend kas nkas as, 
   NoRepeats as,
   AttributesFor kas kts     -- To enforce that elements are actually attributes.
 ) => TABLEHEADER (kas,nkas)

-- | From a table header, we can derive the column types (pair of lists).
class AttributesForPair h ts | h -> ts
instance (AttributesFor kas kts, AttributesFor nkas nkts)
   => AttributesForPair (kas,nkas) (kts,nkts)

-- | From a table header, we can derive the column types (list)
class AttributesForList h l | h -> l
instance (AttributesForPair h (kts,nkts), HAppend kts nkts ts)
   => AttributesForList h ts

-- | A well-formed table has a well-formed header and column types
--   that correspond to the attribute types of that header.
class TABLE t
instance HeaderFor h ks vs  => TABLE (Table h ks vs)

-- | A constructor for table creation that checks well-formedness.
createTable :: (
   TABLE (Table h kts nkts),
   HeaderFor h kts nkts
 ) => h -> Table h kts nkts
createTable h = Table h Map.empty

-----------------------------------------------------------------------------
-- | Find the table row type for a given table.
--   The row type is a record where the columns of
--   the table serve as labels.

class Row h k v r | h -> k v r
 where
  row :: h -> k -> v -> r
  unRow :: h -> r -> (k,v)
instance (
   HeaderFor (a,b) k v,
   HAppend a b ab,
   HAppend k v kv,
   HBreak kv k v,
   HZip ab kv l
 ) => Row (a,b) k v (Record l)
 where
  row (a,b) k v
    = Record $ hZip (hAppend a b) (hAppend k v)
  unRow (a,b) (Record l)
    = hBreak $ snd $ hUnzip l
    
class HBreak l l' l'' | l' l'' -> l 
 where
  hBreak :: l -> (l',l'')
instance HBreak l HNil l
 where
  hBreak l = (HNil,l)
instance (
   HBreak l l' l''
 ) => HBreak (HCons a l) (HCons a l') l''
 where
  hBreak (HCons a l) = (HCons a l',l'') 
   where (l',l'') = hBreak l
 
-- * Representation of query results

-- | Result set, can not contain duplicates.
data (
   AttributesFor a v,
   Ord v
 ) => ResultSet a v = ResultSet a (Set v) 
                            deriving (Eq,Show)
 
class (
   AttributesFor a v,
   Ord v
 ) => RESULTSET a v 
instance (
   AttributesFor a v,
   Ord v
 ) => RESULTSET a v


-- | Result list, can contain duplicates.
data AttributesFor a v => ResultList a v = ResultList a [v] 
                                                deriving (Eq,Show)

data (
      AttributesFor a v, 
      AttributesFor a' k
    ) => ResultGroups a a' k v = ResultGroups a a' (Map k [v])

-- | Transform a result set into a table, where all values are keys.
resultSet2table (ResultSet a s) 
 = Table (a, HNil) (Map.fromList $ map (\x -> (x,HNil)) (Set.toList s))

-----------------------------------------------------------------------------
-- * Defaultable system values
class Defaultable x 
 where
  defaultValue :: x


instance Defaultable Int
 where
  defaultValue = 0

instance Defaultable String
 where
  defaultValue = ""


-- | Tests if an attribute type is null

class IsNull a where isNull :: a -> Bool
instance IsNull (Maybe a) 
 where 
  isNull Nothing  = True
  isNull (Just _) = False
instance IsNull a
 where
  isNull _ = False



-- Seems we need a universal instance for HEq.
instance TypeEq x y b => HEq x y b


