{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances -fallow-undecidable-instances #-}
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  Alexandra Silva, Joost Visser
-- Stability   :  experimental
-- Portability :  portable
--
-- Reconstruction of the Structured Query Language (SQL).
--
-----------------------------------------------------------------------------

module Data.HDB.SQL where

import Data.HDB.Representation hiding (createTable)
import Data.HListExtras
import Data.Map (Map)
import Data.Set (Set)
import qualified Data.Map as Map
import qualified Data.Set as Set
import Prelude hiding (null,product)
import Data.HDB.RelationalAlgebra hiding (filter) 

  
-----------------------------------------------------------------------------
-- * INSERT
--
-- Issues:
--   - multiple vs single row
--   - column list serves as projection
--   - Missing values of nullable types become NULL 

-- | Insert a single row.
--
--     INSERT INTO table-1 [(column-list)] VALUES (value-list)
--


-- | Insert values
   
class InsertValues t as vs where
    insertValues :: t -> as -> vs -> t

instance (
    HeaderFor h k v, 
    Split vs h k v
    ) => InsertValues (Table h k v) HNil vs where 
    insertValues (Table h m) _ vs = Table h m'
        where m' = Map.insert k v m
              (k,v) = split vs h
instance (
    HeaderFor h ks vs,
    HZip (HCons a as) vs' row,
    LineUp h (Record row) (ks,vs) (Record HNil) -- No values are allowed to remain
    ) => InsertValues (Table h ks vs) (HCons a as) vs' where
    insertValues 
     {-into-} (Table h mp) 
     {-[Collumn Spec]-} as 
     {- Values -} vs = Table h mp'
     where
      mp' = Map.insert kvs nkvs mp
      (kvs,nkvs) = fst $ lineUp h (Record $ hZip as vs)

insertValues' :: (
    HeaderFor h ks vs, HZip as vs' row,
    LineUp h (Record row) (ks,vs) (Record HNil)
    ) => (Table h ks vs) -> as -> vs' -> Table h ks vs 
insertValues' (Table h mp) as vs = Table h mp'
     where
      mp' = Map.insert kvs nkvs mp
      (kvs,nkvs) = fst $ lineUp h (Record $ hZip as vs)

insertTable :: (HZip b' v' l',
                LookupMany b l' v,
                LineUp a k' k r,
                LineUpPadd b v' v r',
                HeaderFor (a, b) k v,
                HeaderFor (a', b') k' v') =>
                Table (a', b') k' v'
                -> Table (a, b) k v
                -> Table (a, b) k v
insertTable (Table (a',b') m) (Table (a,b) m') = Table (a,b) m''
    where
        m'' = Map.foldWithKey addRow m' m
        addRow k' v' = Map.insert (fst $ lineUp a k') (fst $ lineUpPadd b v')
        
-- | Insert the result set of a nested query (multiple rows).
--
--     INSERT INTO table-1 [(column-list)] (query-specification)
--
insertResultSet as (ResultSet a v) t
  = Set.fold (\r t -> insertValues t as r) t v

insertResultList :: (
    AttributesFor a v, 
    LineUp h (Record row) (ks, vs) (Record HNil),
    HZip vs' v row,
    HeaderFor h ks vs
 ) => vs' -> ResultList a v -> Table h ks vs -> Table h ks vs
insertResultList as (ResultList a v) t
  = foldr (\vs t -> insertValues' t as vs) t v

-----------------------------------------------------------------------------
-- * UPDATE

-- | Update (multiple rows)
--
--     UPDATE table-1 SET set-list [WHERE predicate]
--   where
--
--  SET column-1 = value-1 [, column-2 = value-2] ...

-- | A set clause is basically a record.
update :: (
   HeaderFor h k v,
   Row h k v r,
   UpdateWith r s
 ) => Table h k v -> s -> (r->Bool) -> Table h k v
update (Table h m) 
  {-set-} s
  {-where-} p
 = Table h m'
  where
    m' = Map.foldWithKey upd Map.empty m
    upd k v
      | p r = Map.insert k' v' 
      | otherwise = Map.insert k v
      where   r = row h k v
              (k',v') = unRow h $ updateWith r s
    
-----------------------------------------------------------------------------
-- * DELETE

-- | Delete (multiple rows)
--   
--    DELETE FROM table-1 [WHERE predicate]
--
delete :: (
   HeaderFor h ks vs,
   Row h ks vs r
 ) => Table h ks vs -> Where r -> Table h ks vs
delete {-from-} t@(Table h m)
       {-where-} p
 = Table h m'
 where
  m' = Map.foldWithKey del Map.empty m
  del k v
    | p (row h k v) = id             
    | otherwise          = Map.insert k v

-----------------------------------------------------------------------------
-- * WHERE

-- | The where clause is a predicate on a single row.
--   The row is represented by a record, for easy access
--   to column values.
type Where r = r -> Bool


-- * SELECT

-- | Flag DISTINCT is just a boolean.
type Distinct = Bool

-- | Select on a single table, without ORDER BY, GROUP BY or HAVING.
select :: (
   HeaderFor h ks vs,
   Row h ks vs r,
   LookupMany a r v,
   AttributesFor a v,
   Ord v           
 ) => Distinct -> a -> Table h ks vs -> Where r -> ResultList a v
select
 {-distinct-} distinct 
              a 
 {-from-}     t@(Table h m)
 {-where-}    p
  = ResultList a results
  where
    fm = Map.filterWithKey (\k v -> p $ row h k v) m
    ps = Map.foldWithKey (\k v m -> Set.insert (lookupMany a (row h k v)) m) Set.empty fm
    pl = Map.foldWithKey (\k v m -> (lookupMany a (row h k v)):m) [] fm
    results = if distinct then Set.toList ps else pl

-- | Via record projection
select'' :: (
   Products ts (Table h ks vs),
   AttributesFor ls a,
   LookupMany ls r a,
   HeaderFor h ks vs,
   Row h ks vs r,
   Ord a
 ) => Distinct -> ls -> ts -> (r -> Bool) -> ResultList ls a
select''
 {-distinct-} distinct 
              a 
 {-from-}     ts
 {-where-}    p
  = ResultList a results
  where
    Table h m = products ts
    fm = Map.filterWithKey (\k v -> p $ row h k v) m
    pl = Map.foldWithKey (\k v m -> lookupMany a (row h k v):m) [] fm
    results = if distinct then rmDbls pl else pl

-- | Via record projection, with order by
-- | select with order by
select''' :: (
   Products ts (Table h k v), HeaderFor h k v,
   Row h k v r, LookupMany a r x, AttributesFor a x, Ord x,
   HZip a x r', LookupMany b (Record r') y, Ord y,
   IsEmpty b
 ) => Bool -> a -> ts -> (r -> Bool) -> b -> ResultList a x
select'''
 {-distinct-} distinct 
              a 
 {-from-}     ts
 {-where-}    p
 {-order by-} b
  = ResultList a $ uniq $ sort $ proj $ fltr m
  where
    Table h m = products ts
    fltr = Map.filterWithKey (\k v -> p $ row h k v)
    proj = Map.foldWithKey flt []
    flt k v l = lookupMany a (row h k v):l
    uniq = if distinct then rmDbls else id
    sort = if isEmpty b then id else (qsort . cmp) b
    cmp b v v' = lkp v < lkp v' 
      where lkp x = lookupMany b (Record $ hZip a x)


-- | function to remove duplicates from a list
rmDbls l = Set.toList $ Set.fromList l


-- | order by for query results
orderBy :: (
  AttributesFor a v,
  HZip a v av,
  LookupMany b (Record av) w, Ord w
 ) => b -> ResultList a v -> ResultList a v
orderBy b (ResultList a rs) 
 = ResultList a $ qsort rowCompare rs
   where
     rowCompare v v' = lookupMany b (Record $ hZip a v) > lookupMany b (Record $ hZip a v')


qsort :: (a -> a -> Bool) -> [a] -> [a]
qsort (<) []     = []
qsort (<) (x:xs) 
  = qsort (<) elts_lt_x ++ [x] ++ qsort (<) elts_greq_x
    where
      elts_lt_x   = [y | y <- xs, y < x]
      elts_greq_x = [y | y <- xs, not $ y < x]
      
staged_compare :: [a -> a -> Bool] -> a -> a -> Bool
staged_compare cs a a' = or $ map (\f -> f a a') cs
-----------------------------------------------------------------------------
-- * Cartesian product

class Product t t' t'' | t t' -> t''
 where
  product :: t -> t' -> t''

-- On tables.
instance (
   HeaderFor (a,b) k v,
   HeaderFor (a',b') k' v',
   HeaderFor (a'',b'') k'' v'',
   HAppend a a' a'',
   HAppend b b' b'',
   HAppend k k' k'', HAppend v v' v''
 ) => Product (Table (a,b) k v) (Table (a',b') k' v') (Table (a'',b'') k'' v'')
 where
  product (Table (a,b) m) (Table (a',b') m') = Table h'' m''
   where
    h'' = (hAppend a a', hAppend b b')
    m'' = product' m m'

-- On ResultLists.
instance (
   AttributesFor a v,
   AttributesFor a' v',
   AttributesFor a'' v'',
   HAppend a a' a'',
   HAppend v v' v''
 ) => Product (ResultList a v) (ResultList a' v') (ResultList a'' v'')
 where
  product (ResultList a v) (ResultList a' v') = ResultList a'' v''
   where
    a'' = hAppend a a'
    v'' = productLists v v'


productLists v v' = [hAppend x y| x <- v, y <-v']

productT :: (
   HeaderFor (a,b) k v,
   HeaderFor (a',b') k' v',
   HAppend v v' v'',
   HAppend k k' k'',
   HAppend b b' b'',
   HAppend a a' a'',
   HeaderFor (a'',b'') k'' v''
 ) => Table (a,b) k v -> Table (a', b') k' v' -> Table (a'', b'') k'' v''
productT (Table (a,b) m) (Table (a',b') m') = Table h'' m''
   where
    h'' = (hAppend a a', hAppend b b')
    m'' = product' m m'

class Products ts t | ts -> t
 where
  products :: ts -> t
instance Products (HCons t HNil) t
 where
  products (HCons t _) = t
instance (
   Products (HCons t' ts) t'',
   Product t t'' t'''
 ) => Products (HCons t (HCons t' ts)) t'''
 where
  products (HCons t ts)
   = product t (products ts) 


product' m m'
   = Map.foldWithKey worker Map.empty m
     where
      worker k v m'' = Map.foldWithKey (\k' v' m'' -> Map.insert (hAppend k k') (hAppend v v') m'') m'' m'
-----------------------------------------------------------------------------
-- * Inner join.

-- | inner join for maps
innerJoin :: (
   Ord k,
   Ord k',
   HAppend k' v' kv',
   HAppend v kv' vkv'
 ) => (k -> v -> k') -> Map k v -> Map k' v' -> Map k vkv'
innerJoin on m m'
   = Map.foldWithKey worker Map.empty m
     where
      worker k v m'' 
        = maybe m'' (\v' -> Map.insert k (hAppend v (hAppend k' v')) m'') (Map.lookup k' m')
          where k' = on k v

-- | Inner Join between two tables with a on clause
sqlInnerJoin :: (
   HeaderFor (a, b) k v,
   HeaderFor (a',b') k' v',
   HeaderFor (a, bab') k vkv',
   HAppend v kv' vkv',
   HAppend k' v' kv',
   HAppend a' b' ab',
   HAppend b ab' bab',
   LookupMany a' r k',
   Row (a,b) k v (Record row)
 ) => Table (a,b) k v -> Table (a',b') k' v' -> (Record row -> r) -> Table (a,bab') k vkv'
sqlInnerJoin t1@(Table h@(a,b) m)
 {-join-} t2@(Table h'@(a',b') m') 
 {-on-}   equi
 = Table h'' m''
 where
   h'' = (a,hAppend b (hAppend a' b'))
   m'' = innerJoin (\k v -> lookupMany a' (equi $ row h k v)) m m'


-- * GROUP BY

-- | for Maps
mgroupbyagregating ats_selector f m = Map.map f $ Map.foldWithKey worker Map.empty m
        where worker k v = Map.insertWith (Map.union) (ats_selector $ hAppend k v) (Map.singleton k v) 

groupByM :: (Ord k', Ord k) => (k -> v -> k') -> (Map k v -> a) -> Map k v -> Map k' a
groupByM key val m = Map.map val $ Map.foldWithKey grp Map.empty m
  where grp k v = Map.insertWith Map.union (key k v) (Map.singleton k v) 

groupByM' ats_selector f m
 = groupByM (ats_selector . hAppend) f m

-- | for tables
selectGrouped
 {-distinct-} distinct 
              a 
 {-from-}     ts
 {-where-}    p
 {-groupby-}  a'
  = ResultGroups a a' results
  where
    Table h@(kats,nkats) m = products ts
    fm = Map.filterWithKey (\k v -> p $ row h k v) m
    groups = groupByM ((lookupMany a') . Record . (hZip (hAppend kats nkats)). hAppend) id fm
    pl = Map.map (Map.foldWithKey (\k v m -> lookupMany a (row h k v):m) []) groups
    results = if distinct then Map.map rmDbls pl else pl

-- Select with Grouping:
--   * aggregate functions are present
--   * having clause is present
--   * groups are collapsed into single rows
-- But:
--   * no projection
{-
selectG :: (
  Products ts (Table h k v), HeaderFor h k v,       -- product
  Row h k v r, LookupMany g r x, Ord x,             -- where, grouping
  HAppend k v kv, HZip x f af, Aggrs h kv f y,      -- flatten, aggregate
  Row a x y r', HAppend x y xy,                     -- having
  HAppend h x a, AttributesFor a xy, Ord xy,        -- listify
  HZip a xy r'', LookupMany b r'' z, Ord z          -- ordering
 ) => Bool -> af -> ts -> (r -> Bool) -> Maybe b -> g -> (r' -> Bool) -> ResultList a xy
-}
selectG distinct f ts p b g q = ResultList h' $ sort $ uniq $ lst gs''
  where
    (l,a) = hUnzip f
    Table h m = products ts
    fm = Map.filterWithKey (\k v -> p $ row h k v) m
    gs = groupByM (\k v -> lookupMany g $ row h k v) flt fm
    flt = Map.foldWithKey (\k v m -> (k,v):m) []
    gs' = Map.map (\kvs -> aggrs h kvs a) gs
    h'  = hAppend g l
    gs'' = Map.filterWithKey (\k v -> q $ row (g,l) k v) gs'
    uniq = if distinct then rmDbls else id
    sort = if isEmpty b then id else (qsort . cmp) b
    cmp b v v' 
     = lookupMany b (hZip h' v) > lookupMany b (hZip h' v')
    lst = Map.foldWithKey (\g kv xs -> (hAppend g kv):xs) []

-- * Aggregate functions

-- | Basically, an aggregate function is an attribute plus a function from a list
--   of values of that attribute to a single value of a new attribute.

-- Example: COUNT(NAME)
-- count  :: [a] -> Int
-- name   :: Attribute String NAME
-- result :: Attribute Int (COUNT NAME)

-- Example: AVG(AGE)
-- avg    :: (Real a, Fractional b) => [a] -> b
-- age    :: Real a => Attribute a AGE
-- result :: Fractional b => Attribute b (AVG,AGE)

data AF r a b = AF ([a] -> b) (r -> a) 

aggregate' :: Row h k v r => AF r a b -> h -> [(k,v)] -> b
aggregate' (AF rdc mp) h kvs 
 = aggr kvs
   where
     aggr kvs = rdc $ map mpRow kvs
     mpRow (k,v) = mp $ row h k v

class Aggrs h kv f x | h kv f -> x where
  aggrs :: h -> [kv] -> f -> x
instance Aggrs h kv HNil HNil where
  aggrs _ _ _ = HNil
instance (
  Aggrs h (k,v) f x, Row h k v r
 ) => Aggrs h (k,v) (HCons (AF r a b) f) (HCons b x) where
  aggrs h kvs (HCons aggr f) = HCons (aggregate' aggr h kvs) (aggrs h kvs f)

sumF e = AF sum e
avgF e = AF avg e
minF e = AF minimum e
maxF e = AF maximum e
countF e = AF (length . filter (==False)) (isNull . e)

avg :: Real a => [a] -> Float
avg xs = (realToFrac $ sum xs) / (fromIntegral $ length xs)






