{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances -fallow-undecidable-instances #-}
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  Alexandra Silva, Joost Visser
-- Stability   :  experimental
-- Portability :  portable
--
-- Unit tests.
--
-----------------------------------------------------------------------------
module Data.HDB.SQLHUnit where

import Test.HUnit

import Data.HDB.SQL 
import Data.HDB.RelationalAlgebra(split)
import Data.HDB.Representation 
import Data.HListExtras

import qualified Data.Map as Map
import Prelude hiding (product)

-----------------------------------------------------------------------------
-- * Setting up test data.

atID    = attribute :: Attribute Int One
atYN    = attribute :: Attribute Bool One
atName  = attribute :: Attribute String One
atAge   = attribute :: Attribute Int Two
atCity  = attribute :: Attribute String Two

keys1 = atID .*. atYN.*. HNil
vals1 = atName .*. atAge .*. atCity .*.HNil

tbl1 = createTable (keys1,vals1)

john = (1::Int) .*. True .*. "John" .*. (25::Int) .*. "London" .*. HNil
jane = (2::Int) .*. True .*. "Jane" .*. (34::Int) .*. "London" .*. HNil
jill = (1::Int) .*. False .*. "Jill" .*. (99::Int) .*. "Paris" .*. HNil

tblWithThree = insertValues (insertValues (insertValues tbl1 HNil jill) HNil jane) HNil john

at2City = attribute :: Attribute String Three
atCountry = attribute :: Attribute String Four

keys2 = at2City .*. HNil
vals2 = atCountry .*. HNil

tbl2 = createTable (keys2,vals2)

londonUK    = "London" .*. "UK" .*. HNil 
parisFrance = "Paris" .*. "France" .*. HNil 
bostonUSA   = "Boston" .*. "USA" .*. HNil

tblCity = insertValues (insertValues (insertValues tbl2 HNil bostonUSA) HNil parisFrance) HNil londonUK 

-----------------------------------------------------------------------------
-- * Main function and auxiliaries

main
  = runTestTT . TestList $ testList
  
labeledTest l t
  = TestLabel l (TestCase t)

tableSize (Table h m) = Map.size m
resultListSize (ResultList h l) = length l

-----------------------------------------------------------------------------
-- * Tests

-- tbl2 _ = insertValues' john tbl1

testList
 = [
    labeledTest "test split keys" $ do
      assertEqual "empty"
        (HNil,HNil) 
        (split HNil (HNil,HNil))
      assertEqual "one key"
        (True .*. HNil,HNil) 
        (split (True .*. HNil) (atYN.*. HNil,HNil))
      assertEqual "one value"
        (HNil,True .*. HNil) 
        (split (True .*. HNil) (HNil,atYN.*. HNil))
      assertEqual "same types for 2 values"
        (HNil,"John" .*. "London" .*. HNil) 
        (split ("John" .*. "London" .*. HNil) 
                   (HNil,atName .*. atCity .*. HNil) )
      assertEqual "same types for key and value"
        ("John" .*. HNil,"London" .*. HNil) 
        (split ("John" .*. "London" .*. HNil) 
                   (atName .*. HNil,atCity .*. HNil) )
      assertEqual "split john"
        ((1::Int) .*. True .*. HNil, "John" .*. (25::Int) .*. "London" .*. HNil) 
        (split ((1::Int) .*. True .*. "John" .*. (25::Int) .*. "London" .*. HNil) 
                   ( atID .*. atYN.*. HNil,
                     atName .*. atAge .*. atCity .*.HNil) )
   {-,
    labeledTest "test insert john" $ do
       assertEqual "self-test" tbl1 tbl1
       let rslt = insertPrimitive 
                     (1 .*. True .*. HNil) 
                     ("John" .*. 25 .*. "London" .*. HNil) tbl1
       assertEqual "insertPrimitive 1 row" 1 (tableSize rslt)
       let rslt' = insertValues
                     tbl1 HNil ((1::Int) .*. True .*. "John" .*. (25::Int) .*. "London" .*. HNil)
       assertEqual "insertValues 1 ordered row" rslt rslt'
       let rslt'' = insertValues
                     tbl1 HNil ("John" .*. (1::Int) .*. (25::Int) .*. "London" .*. True .*. HNil) 
       assertEqual "insertValues 1 row out of order" rslt rslt''
    ,
     labeledTest "test substitute" $ do
       assertEqual "empty" HNil (substitute HNil HNil HNil)
       assertEqual "redudant assign" HNil (substitute HNil HNil ( (True .=. False) .*. HNil ) )
       assertEqual "no assign" 
         (True .*. HNil) 
         (substitute (atYN .*. HNil) (True .*. HNil) HNil)
       assertEqual "ineffective assign" 
         (True .*. HNil) 
         (substitute (atYN .*. HNil) (True .*. HNil) ( (True .=. False) .*. HNil ))
       assertEqual "effective assign" 
         (False .*. HNil) 
         (substitute (atYN .*. HNil) (True .*. HNil) ( (atYN .=. False) .*. HNil ))
       assertEqual "ambiguous assign (uses first)" 
         (False .*. HNil) 
         (substitute (atYN .*. HNil) (True .*. HNil) ( (atYN .=. False) .*. (atYN .=. True) .*. HNil ))
       assertEqual "attributes of same type" 
         ("Paris" .*. "Jane" .*. HNil) 
         (substitute (atCity .*. atName .*. HNil) ("London" .*. "John" .*. HNil) 
                     ( (atCity .=. "Paris") .*. (atName .=. "Jane") .*. HNil ))-}
   ,
    labeledTest "test update all" $ do
       let rslt = update tblWithThree
                    {-set-} HNil
                    {-where-} (const True)
       assertEqual "empty update on all" tblWithThree rslt 
       let rslt = update tblWithThree
                    {-set-} ( atID .=. (0::Int) .*. atYN .=. False .*. HNil )
                    {-where-} (const True)
       assertEqual "all keys become the same" 1 (tableSize rslt) 
   ,
    labeledTest "test update john only" $ do
       let rslt = update tblWithThree
                    {-set-} ( atCity .=. "Amsterdam" .*. HNil )
                    {-where-} (\t -> (t.!.atName) == "John" )
       assertBool "john moves to amsterdam" (tblWithThree /= rslt) 
       let rslt' = update rslt
                    {-set-} ( atCity .=. "London" .*. HNil )
                    {-where-} (\t -> (t.!.atName) == "John" )
       assertEqual "john moves back" tblWithThree rslt' 
   ,
    labeledTest "test delete all" $ do
       let rslt = delete {-from-}  tblWithThree
                         {-where-} (const True)
       assertEqual "everything deleted" tbl1 rslt
   ,
    labeledTest "test delete john" $ do
       let rslt = delete {-from-}  tblWithThree
                         {-where-} (\t -> (t.!.atName) == "John")
       assertEqual "john deleted" 2 (tableSize rslt)
   ,
    labeledTest "test select" $ do
       let rslt = select True
                         (atCity .*. HNil)      
                         {-from-}  tblWithThree
                         {-where-} (const True)
       assertEqual "2 distinct cities" 2 (resultListSize rslt)
       let rslt = select False
                         (atCity .*. HNil)      
                         {-from-}  tblWithThree
                         {-where-} (const True)
       assertEqual "3 cities in total" 3 (resultListSize rslt)
       let rslt = select True
                         (atCity .*. HNil)      
                         {-from-}  tblWithThree
                         {-where-} (\r -> (r.!.atAge) > 65)
       assertEqual "a senior lives in one distinct city" 1 (resultListSize rslt)
       assertEqual "keep only JFK" rl' (select True allsched sched isJFK)
   ,
    labeledTest "test select with ordering" $ do
       let rslt = select''' True
                         (atCity .*. HNil)      
                         {-from-}  (tblWithThree .*. HNil)
                         {-where-} (const True)
                         {-order by-} HNil      
       assertEqual "2 distinct cities" 2 (resultListSize rslt)
       let rslt = select''' False
                         (atCity .*. HNil)      
                         {-from-}  (tblWithThree .*. HNil)
                         {-where-} (const True)
                         {-order by-} HNil
       assertEqual "3 cities in total" 3 (resultListSize rslt)
       let rslt = select''' True
                         (atCity .*. HNil)      
                         {-from-}  (tblWithThree .*. HNil)
                         {-where-} (\r -> (r.!.atAge) > 65)
                         {-order by-} HNil
       assertEqual "a senior lives in one distinct city" 1 (resultListSize rslt)
       assertEqual "keep only JFK" rl' (select True allsched sched isJFK)
       let rslt = select''' False
                         (atCity .*. HNil)      
                         {-from-}  (tblWithThree .*. HNil)
                         {-where-} (const True)
                         {-order by-} (atCity .*. HNil)
       assertEqual "3 cities in total" 3 (resultListSize rslt)
       let (ResultList _ vs) = rslt
       assertEqual "Right order?" [HCons "London" HNil,HCons "London" HNil,HCons "Paris" HNil] vs
   ,
    labeledTest "test cartesian product on maps" $ do
       let e = Map.fromList [(HNil,HNil)]
       assertEqual "product of empties is empty" e (product' e e)
       let m = Map.fromList [("Aap" .*. HNil,HNil),("Noot" .*. HNil,HNil),("Mies" .*. HNil,HNil)]
       assertEqual "product with empty" m (product' e m)
       let m' = Map.fromList [(True .*. HNil,HNil),(False .*. HNil,HNil)]
       assertEqual "product 3 x 2 = 6" 6 (Map.size $ product' m m')
       assertEqual "product 3 x 3 = 9" 9 (Map.size $ product' m m)
   ,
    labeledTest "test multi-table select" $ do
       let rslt = select'' False
                          (atName .*. atCountry .*. HNil)
                          {-from-}  (tblWithThree .*. tblCity .*. HNil)
                          {-where-} (\r -> (r.!.atCity) == (r.!.at2City))
       assertEqual "three people" 3 (resultListSize rslt)
       let rslt = select'' True
                          (atCountry .*. HNil)
                          {-from-}  (tblWithThree .*. tblCity .*. HNil)
                          {-where-} (\r -> (r.!.atCity) == (r.!.at2City))
       assertEqual "two inhabited countries" 2 (resultListSize rslt)
   ,
    labeledTest "test inner join" $ do
       let rslt = select  False
                          (atName .*. atCountry .*. HNil)
                          {-from-}  ( sqlInnerJoin tblWithThree tblCity 
                                      {-on-} (\r -> ((at2City .=. (r.!.atCity)) .*. HNil)) )
                          {-where-} (const True)
       assertEqual "three people" 3 (resultListSize rslt)

 ]


-- Maier page 14

data NUMB; instance Show NUMB where show _ = "NUMB"
data FROM; instance Show FROM where show _ = "FROM"
data TO; instance Show TO where show _ = "TO"
data DEPARTS; instance Show DEPARTS where show _ = "DEPARTS"
data ARRIVES; instance Show ARRIVES where show _ = "DEPARTS"

emptyscheduleTable = createTable hsched

hsched = ((attribute :: Attribute Int NUMB)   .*. HNil
 ,
  (attribute :: Attribute String FROM)      .*.
  (attribute :: Attribute String TO)        .*. 
  (attribute :: Attribute String DEPARTS)   .*. 
  (attribute :: Attribute String ARRIVES)   .*. 
  HNil
 )

allsched = uncurry hAppend hsched

f1 = (84::Int) .*. "O'Hare" .*. "JFK" .*. "3:00p" .*. "5:55p" .*. HNil
f2 = (109::Int) .*. "JFK" .*. "Los Angeles" .*. "9:40p" .*. "2:42a" .*. HNil
f3 = (117::Int) .*. "Atlanta" .*. "Boston" .*. "10:05p" .*. "12:43a" .*. HNil
f4 = (213::Int) .*. "JFK" .*. "Boston" .*. "11:43a" .*. "12:45p" .*. HNil
f5 = (214::Int) .*. "Boston" .*. "JFK" .*. "2:20p" .*. "3:12p" .*. HNil
rl = ResultList allsched [f1,f2,f3,f4,f5]
sched = insertResultList allsched rl emptyscheduleTable 

rl' = ResultList allsched [f2,f4]
sched' = insertResultList allsched rl' emptyscheduleTable 

atFrom    = attribute :: Attribute String FROM

isJFK :: (HasField (Attribute String FROM) r [Char]) => r -> Bool
isJFK = (\r -> (r.!.atFrom) == "JFK")
-----------------------------------------------------------------------------
