-----------------------------------------------------------------------------
-- |
-- Maintainer  :  Joost Visser and Cupertino Miranda
-- Stability   :  experimental
-- Portability :  portable
--
-- Derive data flow graph for spreadsheets.
--
-----------------------------------------------------------------------------

module Language.Gnumeric.CellBlock where

import Language.Gnumeric.Syntax
import Language.Gnumeric.Formula

-----------------------------------------------------------------------------

-- | Type for block of cells. First cell reference is the upper-left corner 
--   of the block. The lower-left corner is either the same cell reference,
--   or the second one if it exists.
type  CellBlock = (CellRef, Maybe CellRef)

-- | Constructor to CellRef
mkCellRef :: String -> Gmr'Cell_Attrs -> CellRef
mkCellRef sheetName attrs
        = CellRef { 
	    sheet = sheetName,
            colNr = read $ gmr'CellCol attrs, 
            rowNr = read $ gmr'CellRow attrs
          }

-- | Show function to CellBlock
showCellBlock :: CellBlock -> String
showCellBlock (cr,Nothing)  = showCellRef cr
showCellBlock (cr,Just cr') = (showCellRef cr)++":"++(fromColNr $ colNr cr')++(fromRowNr $ rowNr cr')


-- | Expand a CellBlock to list of CellRefs
expandBlock :: CellBlock -> [CellRef]
expandBlock (c1, Just c2) = [(CellRef (sheet c1) b c) | b <- [(colNr c1)..(colNr c2)], c <- [(rowNr c1)..(rowNr c2)]]
expandBlock (c1,  Nothing) = [c1] 


-- | Compare two CellBlocks
compareCellBlocks :: CellBlock -> CellBlock -> Maybe Ordering
compareCellBlocks (cell1, Nothing) cellBlock
  = compareCellBlocks (cell1, Just cell1) cellBlock
compareCellBlocks cellBlock (cell2, Nothing)
  = compareCellBlocks cellBlock (cell2, Just cell2)
compareCellBlocks (cell1,Just cell2) (cell3,Just cell4)
  = case (cell3 `compare` cell1, cell2 `compare` cell4) of
      (Nothing,_) -> Nothing
      (_,Nothing) -> Nothing
      (Just x, Just y) -> oi x y
    where
      compare = compareCellRefs

-- | Not named function
oi x y
  = case (x,y) of
      (GT,LT) -> Nothing
      (LT,GT) -> Nothing
      (EQ,EQ) -> Just EQ
      (LT,LT) -> Just LT
      (EQ,LT) -> Just LT
      (LT,EQ) -> Just LT
      (GT,GT) -> Just GT
      (EQ,GT) -> Just GT
      (GT,EQ) -> Just GT

-- | Compare CellRefs
compareCellRefs :: CellRef -> CellRef -> Maybe Ordering
compareCellRefs cell1 cell2
  = case (r1 `compare` r2, c1 `compare` c2) of
      (GT,LT) -> Nothing
      (LT,GT) -> Nothing
      (EQ,EQ) -> Just EQ
      (LT,LT) -> Just LT
      (EQ,LT) -> Just LT
      (LT,EQ) -> Just LT
      (GT,GT) -> Just GT
      (EQ,GT) -> Just GT
      (GT,EQ) -> Just GT
  where
    r1 = rowNr cell1
    c1 = colNr cell1
    r2 = rowNr cell2
    c2 = colNr cell2
    (r,c) = (r1 `compare` r2, c1 `compare` c2)
