-----------------------------------------------------------------------------
-- |
-- Maintainer  :  Cupertino Miranda and Joost Visser and Alexandra Silva
-- Stability   :  experimental
-- Portability :  portable
--
-- Derive data flow graph for spreadsheets. Two variations of data flow
-- graphs are supported. The first variation has nodes of type CellBlock.
-- In the second variations, cell blocks are expanded to their constituent
-- cells, and the node type is CellRef. The first variation is more concise
-- and more suitable for human consumption. The second variation is more
-- suited for spreadsheet manipulation, such as slicing.
--
-----------------------------------------------------------------------------

module Language.Gnumeric.Dataflow where

import Data.Graph.Basics 
import Language.Gnumeric.Syntax
import Language.Gnumeric.SyntaxTermInstances()
import Language.Gnumeric.Formula
import Language.Gnumeric.FormulaTermInstances()
import Language.Gnumeric.CellBlock
import qualified Data.Set as Set

import StrategyLib
import Control.Monad.Identity
import Data.Monoid

-----------------------------------------------------------------------------
-- * Two variations of data flow extraction.

-- | Create a data flow graph from a Workbook (spreadsheet), where the nodes
--   of the graph are cell blocks.

dataflow :: Term a => a -> Gph CellBlock
dataflow a
  = flow' `Set.union` nest
    where
      flow = dataflowEdgesSheet (dataflowEdgesCell collectBlockRefs) a
      flow' = Set.map (\(c,b) -> ((c,Nothing),b)) flow
      nest = nesting $ Set.toList $ ent flow'
      
-- | Create a data flow graph from a Workbook (spreadsheet), where the nodes
--   of the graph are individual cell references. It is similar to 'dataflow'
--   except that cell blocks are expanded into their constituent cells.

dataflowExpanded :: Term a => a -> Gph CellRef
dataflowExpanded a
  = dataflowEdgesSheet (dataflowEdgesCell collectCellRefs) a
  
-----------------------------------------------------------------------------
-- * General algorithm for data flow extraction.

-- | Construct dataflow edges for a complete sheet.
dataflowEdgesSheet :: (Term a, Ord b) 
                   => (forall a. Term a => String -> a -> Rel CellRef b)
                   -> a 
                   -> Rel CellRef b 
dataflowEdgesSheet collector a
  = maybe mempty id $ 
      applyTU (stop_tdTU (failTU `adhocTU` worker)) a
    where
      --worker :: Ord b => Gmr'Sheet -> Maybe (Rel CellRef b)
      worker (Gmr'Sheet _ (Gmr'Name name) _ _ _ _ _ _ _ _ _ _ cells _ _ _ _)
        = Just $ collector name cells

-- | Create CellBlock edges from all Gmr'Cells
dataflowEdgesCell :: (Term a, Ord b) 
                  => (forall a . Term a => a -> [b])
                  -> String 
                  -> a 
                  -> Rel CellRef b
dataflowEdgesCell collector sheetName a
  = maybe mempty id $ 
      applyTU (crush (constTU emptyRel `adhocTU` worker)) a
    where
      --worker :: Gmr'Cell -> Maybe (Rel CellRef CellBlock)
      worker cell = Just $ dataflowEdgesFormula collector sheetName cell

-- | Create data flow edges for a given cell given a Gmr'Cell collector
dataflowEdgesFormula :: (Ord b) 
                     => (forall a . Term a => a -> [b])
                     -> String 
                     -> Gmr'Cell 
                     -> Rel CellRef b
dataflowEdgesFormula collector sheetName (Gmr'Cell attrs content) 
  = mkRel [ (cellRef,y) | y <- cellRefs ] 
    where
      cellRefs = collector (gmr'CellAst attrs)
      cellRef = mkCellRef sheetName attrs

-- | Create nesting relationships between cell blocks and their individual
--   constituent cells. Only cells that appear themselves among the
--   given blocks are considered.
nesting :: [CellBlock] -> Gph CellBlock
nesting bs
  = mkRel $ concatMap worker bs
    where
      worker b 
        = map (\x -> (x,b)) (filter (isInside b) bs)
      b1 `isInside` b2 
        = b1 `compareCellBlocks` b2 == Just LT

-----------------------------------------------------------------------------
-- * Functions to instantiate the general algorithm

-- | Collect all block references from a given term.
collectBlockRefs :: Term a => a -> [CellBlock]
collectBlockRefs a
  = runIdentity $ applyTU (collect (constTU [] `adhocTU` worker)) a
    where
      worker :: Formula -> Identity [CellBlock]
      worker (Ref c1 c2) = return [(c1,c2)]
      worker _ = return []

-- | Collect all cell references from a given term.
--   Blocks are expanded to to their constituent CellRefs.
collectCellRefs :: Term a => a -> [CellRef]
collectCellRefs a
  = runIdentity $ applyTU (collect (constTU [] `adhocTU` worker)) a
    where
      worker :: Formula -> Identity [CellRef]
      worker (Ref c1 c2) = return $ expandBlock $ (c1,c2)
      worker _ = return []
   
-----------------------------------------------------------------------------

{- Now offered by Data.Set
-- | Definition of Set as instance of Monoid
instance Ord a => Monoid (Set a) where
  mempty = emptySet
  mappend = union
-}

-----------------------------------------------------------------------------
