-----------------------------------------------------------------------------
-- |
-- Maintainer  :  Joost Visser and Cupertino Miranda and Alexandra Silva
-- Stability   :  experimental
-- Portability :  portable
--
-- Filter functions for cells and cell blocks in workbooks 
--
-----------------------------------------------------------------------------

module Language.Gnumeric.Filter where

import Data.Relation.SetOfPairs 
import Data.Set(toList)
import Language.Gnumeric.Syntax
import Language.Gnumeric.SyntaxTermInstances()
import Language.Gnumeric.Formula
import Language.Gnumeric.FormulaTermInstances()
import Language.Gnumeric.CellBlock

import StrategyLib
import Control.Monad.Identity

-----------------------------------------------------------------------------

-- |  Filter the elements in the workbook that are present in the set passed as first argument

filterExpandedDataflow :: Set CellRef -> Gmr'Workbook -> Gmr'Workbook
filterExpandedDataflow cellSet workbook 
  = maybe workbook id $
      applyTP (stop_tdTP (failTP `adhocTP` worker)) workbook
    where
      worker :: Gmr'Sheet -> Maybe (Gmr'Sheet)
      worker sheet@(Gmr'Sheet _ (Gmr'Name name) _ _ _ _ _ _ _ _ _ _ cells _ _ _ _)
        = Just $ replaceCells (filterCellRefsInSheet name cellSet cells) sheet

-- | Filter the elements in the workbook that are present in the set passed as first argument

filterCells :: Set CellBlock -> Gmr'Workbook -> Gmr'Workbook
filterCells cellSet workbook 
  = maybe workbook id $
      applyTP (stop_tdTP (failTP `adhocTP` worker)) workbook
    where
      worker :: Gmr'Sheet -> Maybe (Gmr'Sheet)
      worker sheet@(Gmr'Sheet _ (Gmr'Name name) _ _ _ _ _ _ _ _ _ _ cells _ _ _ _)
        = Just $ replaceCells (filterBlockCellsInSheet name cellSet cells) sheet

-- | Replace element cells from Gmr'Sheet

replaceCells cells (Gmr'Sheet _1 (Gmr'Name name) _2 _3 _4 _5 _6 _7 _8 _9 _10 _11 _ _12 _13 _14 _15) =
  (Gmr'Sheet _1 (Gmr'Name name) _2 _3 _4 _5 _6 _7 _8 _9 _10 _11 cells _12 _13 _14 _15)

-- | Filter the CellBlocks in the worksheet that are members of the set passed as argument
--   The first argument is the sheet name needed for cellRef comparison

filterBlockCellsInSheet :: Term a => String -> Set CellBlock -> a -> a
filterBlockCellsInSheet sheetName cells a@worksheet
  = runIdentity $ applyTP (full_tdTP (idTP `adhocTP` (return . worker))) a
    where
      worker :: Gmr'Cell -> Gmr'Cell
      worker cell 
        | needsRemoval cell = emptyCell cell
        | otherwise = cell  
      needsRemoval :: Gmr'Cell -> Bool
      needsRemoval (Gmr'Cell attrs _) =
           or $ map (cellRef `isInsideEq`)  $ toList cells
         where
           cellRef = mkCellRef sheetName attrs 
      cr `isInsideEq` b2
        = comp == Just LT || comp == Just EQ 
          where comp = (cr, Nothing) `compareCellBlocks` b2

-- | Filter the CellRefs in the worksheet that are members of the set passed as argument
--   The first argument is the sheet name needed for cellRef comparison


filterCellRefsInSheet :: Term a => String -> Set CellRef -> a -> a
filterCellRefsInSheet sheetName cells a@worksheet
  = runIdentity $ applyTP (full_tdTP (idTP `adhocTP` (return . worker))) a
    where
      worker :: Gmr'Cell -> Gmr'Cell
      worker cell 
        | doesNotNeedsRemoval cell = cell
        | otherwise = emptyCell cell  
      doesNotNeedsRemoval :: Gmr'Cell -> Bool
      doesNotNeedsRemoval (Gmr'Cell attrs _) =
           or $ map (cellRef ==)  $ toList cells
         where
           cellRef = mkCellRef sheetName attrs 

-- | Create empty Gmr'Cell

emptyCell :: Gmr'Cell -> Gmr'Cell
emptyCell (Gmr'Cell atrs _) = Gmr'Cell atrs' []
    where
      atrs' = atrs 
	    { gmr'CellExprID = Nothing
	    , gmr'CellValueType = Nothing
	    , gmr'CellValueFormat = Nothing
	    , gmr'CellCols = Nothing
	    , gmr'CellRows = Nothing
	    , gmr'CellAst = Nothing
	    }
