-----------------------------------------------------------------------------
-- |
-- Maintainer  :  Cupertino Miranda and Joost Visser and Alexandra Silva
-- Stability   :  experimental
-- Portability :  portable
--
-- Syntax for spreadsheet formulas.
--
-----------------------------------------------------------------------------

module Language.Gnumeric.Formula where

import Text.ParserCombinators.Parsec
import qualified Text.ParserCombinators.Parsec.Token as P
import Text.ParserCombinators.Parsec.Language
import Text.ParserCombinators.Parsec.Expr

import Data.List (elemIndices)

-----------------------------------------------------------------------------
-- * Abstract syntax

-- | The type of formulas.
data Formula 
  = Var String                      -- ^ Variables
  | Fun String [Formula]            -- ^ Function applications
  | Ref CellRef (Maybe CellRef)     -- ^ Cell or cell block reference
  deriving (Eq,Show,Read)

-- | Type of cell references
data CellRef 
  = CellRef {
      sheet :: String,     -- ^ Sheet name
      colNr :: Int,        -- ^ Column index
      rowNr :: Int         -- ^ Row index
    }
    deriving (Eq,Read,Ord)


parseCellRef :: String -> String -> CellRef
parseCellRef sourceName s
  = case (parse sheetCellReference sourceName s) of
      (Left parseError)
        -> error $ show parseError
      (Right cellref)
        ->  cellref


data ShowReadCellRef 
  = ShowReadCellRef {
      cellref :: CellRef 
    }
    deriving (Eq, Ord)

instance Show CellRef where
    showsPrec _ cell = showString $ showCellRef $ cell

instance Show ShowReadCellRef where
    showsPrec _ cell = showString $ showCellRef $ cellref cell

instance Read ShowReadCellRef where
    readsPrec _ str = [(ShowReadCellRef (parseCellRef "" str),str)]
	

-- | Print a cell reference to a String.
showCellRef :: CellRef -> String
showCellRef cr              = (sheet cr)++"!"++(fromColNr $ colNr cr)++(fromRowNr $ rowNr cr)



-- | Add sheetname to all CellRef elements on Formula
addSheetName :: String -> Formula -> Formula
addSheetName name (Var a) = (Var a)
addSheetName name (Fun s formulas) = Fun s (map (addSheetName name) formulas)
addSheetName name (Ref (a) Nothing) = Ref (addSheetNameCellRef name a) Nothing
addSheetName name (Ref (a) (Just b)) = Ref (addSheetNameCellRef name a) (Just (addSheetNameCellRef name b))

-- | Add sheet name to CellRef
addSheetNameCellRef name a 
  = case (sheet a) of
    "" -> CellRef {sheet=( name), colNr=colNr a, rowNr=rowNr a}
    otherwise -> a

-----------------------------------------------------------------------------
-- * Parsers, based on Parsec.

-- | Top-level invocation of parser.
parseFormula :: String -> String -> String -> Formula
parseFormula sheetName sourceName s
  = case (parse formula sourceName s) of
      (Left parseError)
	-> Var s
      (Right formula)
        -> addSheetName sheetName formula

-- | Formulas.
formula :: Parser Formula
formula = do
  (symbol "=")
  expr

-- | Expressions.
expr :: Parser Formula
expr    =   buildExpressionParser table term
        <?> "expression"

-- | Table to build expression grammar.        
table :: OperatorTable Char () Formula
table   = [ [prefix "-", prefix "+"]
          , [postfix "++"]
          , [binary "*" AssocLeft, binary "/"  AssocLeft ]
          , [binary "+" AssocLeft, binary "-"  AssocLeft ]
          , [binary "=" AssocLeft, binary "<="  AssocLeft ]
          ]

-- | Helper.
binary :: String -> Assoc -> Operator Char () Formula        
binary name assoc 
  = Infix (do{ reservedOp name; return $ \x y -> Fun name [x,y] }) assoc

-- | Helper.
prefix :: String -> Operator Char () Formula        
prefix name       
  = Prefix (do{ reservedOp name; return $ \x -> Fun name [x] })

-- | Helper.
postfix :: String -> Operator Char () Formula        
postfix name       
  = Postfix (do{ reservedOp name; return $ \x -> Fun name [x] })

-- | Terms (leafs of expressions).
term :: Parser Formula
term    =   parens expr 
        <|> do { n <- stringLiteral; return $ Var n }
        <|> try floating
        <|> do { n <- natural; return $ Var (show n) }
        <|> try cellReference
        <|> functionApplication
        <?> "simple expression"

floating :: Parser Formula
floating = do {
    n <- natural;
    char '.';
    n1 <- natural;
    return $ Var ((show n) ++ "." ++ (show n1)) }

word :: Parser String
word = many1 (letter <|> digit)

separator :: Parser ()
separator = skipMany1 (space)

parseTest = parse

sheetCellReference :: Parser CellRef
sheetCellReference = do {
  sheetName <- sepBy1 word separator;
  char '!' ;
  ls <- many1 upper;
  ds <- many1 digit;
  return $ CellRef {sheet=(head sheetName), colNr=toColNr ls, rowNr=toRowNr ds}
}

singleCellReference :: Parser CellRef
singleCellReference = do
  ls <- many1 upper
  ds  <- many1 digit
  return $ CellRef {sheet="", colNr=toColNr ls, rowNr=toRowNr ds}

cellReference :: Parser Formula
cellReference = do 
  c1 <- (try (sheetCellReference) <|> singleCellReference) 
  c2 <- optionMaybe $ do
    char ':'
    singleCellReference
  return $ Ref c1 c2
  
-- | Function application.
functionApplication :: Parser Formula
functionApplication = do
  i <- identifier
  fs <- option [] (parens (sepBy expr (symbol ",")))
  return $ Fun i fs

optionMaybe p
  = option Nothing (do{a <- p; return $ Just a})
  
-----------------------------------------------------------------------------
-- * Lexical issues.

lexer :: P.TokenParser ()
lexer  = P.makeTokenParser gnumericStyle

gnumericStyle
 = LanguageDef 
   { commentStart   = ""
   , commentEnd     = ""
   , commentLine    = ""
   , nestedComments = True
   , identStart     = letter <|> char '_' <|> char '#'
   , identLetter    = alphaNum <|> oneOf "_':!"
   , opStart        = opLetter emptyDef
   , opLetter       = oneOf ":!#$%&*+./<=>?@\\^|-~"
   , reservedOpNames= []
   , reservedNames  = []
   , caseSensitive  = True
   }

whiteSpace= P.whiteSpace lexer
lexeme    = P.lexeme lexer
symbol    = P.symbol lexer
natural   = P.natural lexer
parens    = P.parens lexer
semi      = P.semi lexer
identifier= P.identifier lexer
reserved  = P.reserved lexer
reservedOp= P.reservedOp lexer
stringLiteral = P.stringLiteral lexer

-----------------------------------------------------------------------------

-- | Infinite sequence of letter combinations that identify spreadsheet columns.
columns :: [String]  
columns 
  = [ xs++[x] | xs <- "":columns, x <- ['A'..'Z'] ]

-- | Convert a column name into a column number. For instance, "A" becomes
--   0 and "AA" becomes 26.
toColNr :: String -> Int
toColNr colName
  = head (elemIndices colName columns)

-- | Convert a row name into a row number. For instance, "1" becomes
--   0 and "10" becomes 9.
toRowNr :: String -> Int
toRowNr rowName
  = (read rowName)-1
  
-- | Convert a column number into a column name. Inverse of 'toColNr'.
fromColNr :: Int -> String
fromColNr colNr
  = columns!!colNr

-- | Convert a row number into a row name. Inverse of 'toRowNr'.
fromRowNr :: Int -> String
fromRowNr rowNr
  = show (rowNr+1)
-----------------------------------------------------------------------------
