-----------------------------------------------------------------------------
-- |
-- Maintainer  :  Joost Visser and Cupertino Miranda
-- Stability   :  experimental
-- Portability :  portable
--
-- This module implements a form of type inference on spread sheets. The
-- idea is to to group spreadsheet variables that are related through
-- formulas into a type graph, where nodes are sets of variables, and
-- edges represent subtype relations. A generic engine is used for type
-- inference, see Data.Relation.TypeInference. This engine is fed with
-- simple relations between variables computed in this module.
--
-----------------------------------------------------------------------------

module Language.Gnumeric.TypeInference where

import Language.Gnumeric.Syntax
import Language.Gnumeric.SyntaxTermInstances()
import Language.Gnumeric.Formula
import Language.Gnumeric.FormulaTermInstances()
import Language.Gnumeric.CellBlock
import Language.Gnumeric.Dataflow()

import Data.Graph.Basics 
import Data.Set(unions)

import StrategyLib
import Control.Monad.Identity()
import Data.Monoid

-----------------------------------------------------------------------------
-- * Representation

-- | The types being infered are (groups of) cell references.
type Type = CellRef

-- | Cell block plays the role of array types (not used currently).
type ArrayType = CellBlock

-- | create CellBlock edges from all Gmr'Cells


typeInferenceSheet :: (Term a)
                      => (Type -> Formula -> Gph Type)
                      -> a
                      -> Gph Type
typeInferenceSheet infer a
  = maybe mempty id $ 
      applyTU (stop_tdTU (failTU `adhocTU` worker)) a
    where
      worker (Gmr'Sheet _ (Gmr'Name name) _ _ _ _ _ _ _ _ _ _ cells _ _ _ _)
        = Just $ typeInferenceCells infer name cells


typeInferenceCells :: (Term a) 
                   => (Type -> Formula -> Gph Type)
                   -> String
                   -> a 
                   -> Gph Type
typeInferenceCells infer name a
  = maybe mempty id $ 
      applyTU (crush (constTU emptyRel `adhocTU` worker)) a
    where
      worker :: Gmr'Cell -> Maybe (Gph Type)
      worker (Gmr'Cell attrs _) = Just $ infer' attrs $ gmr'CellAst attrs
      infer' :: Gmr'Cell_Attrs -> Maybe Formula -> Gph Type
      infer' attrs (Just a) = infer (mkCellRef name attrs) a 
      infer' attrs (Nothing) = emptyRel

-----------------------------------------------------------------------------
-- * Extraction of variable graph

-- | Infer type relationships from a formula, given the type of the cell
--   in which it appears.
			  --Rel CellRef CellRef

-- | Collect all cell references from a given term.
--   Blocks are expanded to to their constituent CellRefs.
inferExpression :: Type -> Formula -> Gph Type
inferExpression t (Var str)               = emptyRel
inferExpression t (Fun "sum" fms)         = unions vGs
  where vGs = map (inferExpression t) fms
inferExpression t (Fun "if" [c,thn,lse])  = unions $ vGs ++ [vOp]
  where vGs = (map (inferExpression t) [thn,lse])
        vOp = relateAllInSet (inferExpression' c)
inferExpression t (Fun "+" [a,b]) = unions vGs
  where vGs = map (inferExpression t) [a,b]
inferExpression t (Fun "-" [a,b]) = unions vGs
  where vGs = map (inferExpression t) [a,b]
inferExpression t (Fun "=" [a,b]) = unions vGs
  where vGs = map (inferExpression t) [a,b]
inferExpression t (Fun str fms)           = error $ "Unhandled function: "++str++
                                          " with "++(show $ length fms)++
                                          " arguments."
inferExpression t (Ref c1 c2)             = mkRel [ (t,c) | c <- cs ]
  where cs = expandBlock $ (c1,c2)
  
relateAllInSet ::  (Ord a) => [a] -> Rel a a
relateAllInSet a = mkRel $ zip a (tail a) 

inferExpression' :: Formula -> [Type]
inferExpression' (Fun _ fms)     = concatMap inferExpression' fms
inferExpression' (Ref c1 c2)     = expandBlock $ (c1,c2)
inferExpression' _ = []


inferAttr :: Type -> Formula -> Gph Type
inferAttr t (Var str)         = emptyRel
inferAttr t (Fun "*" fms)     = unions vGs
  where vGs = map (inferAttr t) fms
inferAttr t (Fun "/" fms)     = unions vGs
  where vGs = map (inferAttr t) fms
inferAttr t (Fun "if" [c, thn,lse])  = unions $ vGs 
  where vGs = (map (inferAttr t) [thn,lse])
inferAttr t (Ref c1 c2)             = mkRel [ (t,c) | c <- cs ]
  where cs = expandBlock $ (c1,c2)
inferAttr _ _                 = emptyRel

